from unittest.mock import AsyncMock, MagicMock, call, patch

import pytest

from flama import endpoints, exceptions, http, types
from flama.applications import Flama
from flama.routing.routes.base import BaseRoute
from flama.routing.routes.http import BaseHTTPEndpointWrapper, HTTPEndpointWrapper, HTTPFunctionWrapper, Route


class TestCaseBaseHTTPEndpointWrapper:
    @pytest.fixture(scope="function")
    def handler(self, request):
        async def foo():
            return {"foo": "bar"}

        return foo

    @pytest.fixture(scope="function")
    def wrapper_cls(self):
        class _Wrapper(BaseHTTPEndpointWrapper):
            async def __call__(self, scope: types.Scope, receive: types.Receive, send: types.Send) -> None: ...

        return _Wrapper

    @pytest.fixture(scope="function")
    def endpoint(self, handler, wrapper_cls):
        return wrapper_cls(handler)

    @pytest.mark.parametrize(
        ["response", "result"],
        [
            pytest.param({"foo": "bar"}, http.APIResponse(content={"foo": "bar"}), id="dict"),
            pytest.param(["foo"], http.APIResponse(content=["foo"]), id="list"),
            pytest.param("foo", http.APIResponse(content="foo"), id="str"),
            pytest.param(b"foo", http.APIResponse(content=b"foo"), id="bytes"),
            pytest.param(http.Response(content=b"foo"), http.Response(content=b"foo"), id="response"),
            pytest.param(None, http.APIResponse(content=""), id="none"),
        ],
    )
    def test_build_api_response(self, endpoint, response, result):
        assert endpoint._build_api_response(response) == result


class TestCaseHTTPFunctionWrapper:
    @pytest.fixture(scope="function")
    def app(self):
        app = MagicMock(spec=Flama)
        app.router = MagicMock(resolve_route=MagicMock(side_effect=lambda x: (AsyncMock(), x)))
        app.injector = MagicMock(inject=AsyncMock(side_effect=lambda x, y: x))
        return app

    @pytest.fixture(scope="function")
    def endpoint(self):
        async def foo():
            return {"foo": "bar"}

        return HTTPFunctionWrapper(foo)

    async def test_call(self, app, endpoint):
        scope, receive, send = (
            types.Scope({"app": app, "path": "/", "root_app": app, "type": "http", "method": "GET"}),
            AsyncMock(),
            AsyncMock(),
        )

        await endpoint(scope, receive, send)

        assert receive.call_args_list == []
        assert send.call_args_list == [
            call(
                {
                    "type": "http.response.start",
                    "status": 200,
                    "headers": [(b"content-length", b"13"), (b"content-type", b"application/json")],
                }
            ),
            call({"type": "http.response.body", "body": b'{"foo":"bar"}'}),
        ]


class TestCaseHTTPEndpointWrapper:
    @pytest.fixture(scope="function")
    def app(self):
        app = MagicMock(spec=Flama)
        app.router = MagicMock(resolve_route=MagicMock(side_effect=lambda x: (AsyncMock(), x)))
        app.injector = MagicMock(inject=AsyncMock(side_effect=lambda x, y: x))
        return app

    @pytest.fixture(scope="function")
    def endpoint(self):
        class Endpoint(endpoints.HTTPEndpoint):
            async def get(self):
                return {"foo": "bar"}

        return HTTPEndpointWrapper(Endpoint)

    async def test_call(self, app, endpoint):
        scope, receive, send = (
            types.Scope({"app": app, "path": "/", "root_app": app, "type": "http", "method": "GET"}),
            AsyncMock(),
            AsyncMock(),
        )

        await endpoint(scope, receive, send)

        assert receive.call_args_list == []
        assert send.call_args_list == [
            call(
                {
                    "type": "http.response.start",
                    "status": 200,
                    "headers": [(b"content-length", b"13"), (b"content-type", b"application/json")],
                }
            ),
            call({"type": "http.response.body", "body": b'{"foo":"bar"}'}),
        ]


class TestCaseRoute:
    @pytest.fixture(scope="function")
    def route(self):
        return Route("/", lambda: None, name="foo", include_in_schema=False)

    @pytest.fixture(scope="function")
    def endpoint(self, request):
        if request.param == "function":

            def foo(): ...

            return foo

        elif request.param == "endpoint":

            class FooEndpoint(endpoints.HTTPEndpoint):
                def get(self): ...

                def post(self): ...

            return FooEndpoint
        elif request.param is None:
            return None
        else:
            raise ValueError("Wrong value")

    @pytest.mark.parametrize(
        ["endpoint", "name", "wrapper", "methods", "expected_methods", "exception"],
        (
            pytest.param(
                "function",
                "foo",
                HTTPFunctionWrapper,
                {"GET", "POST"},
                {"GET", "POST", "HEAD"},
                None,
                id="function_explicit_methods",
            ),
            pytest.param(
                "function",
                "foo",
                HTTPFunctionWrapper,
                {"POST"},
                {"POST"},
                None,
                id="function_no_get_method",
            ),
            pytest.param(
                "function",
                "foo",
                HTTPFunctionWrapper,
                None,
                {"GET", "HEAD"},
                None,
                id="function_no_methods",
            ),
            pytest.param(
                "endpoint",
                "FooEndpoint",
                HTTPEndpointWrapper,
                {"GET", "POST"},
                {"GET", "POST", "HEAD"},
                None,
                id="endpoint_explicit_methods",
            ),
            pytest.param(
                "endpoint",
                "FooEndpoint",
                HTTPEndpointWrapper,
                {"POST"},
                {"POST"},
                None,
                id="endpoint_no_get_method",
            ),
            pytest.param(
                "endpoint",
                "FooEndpoint",
                HTTPEndpointWrapper,
                None,
                {"GET", "POST", "HEAD"},
                None,
                id="endpoint_no_methods",
            ),
            pytest.param(
                None,
                None,
                None,
                None,
                None,
                exceptions.ApplicationError("Endpoint must be a callable or an HTTPEndpoint subclass"),
                id="error_handler",
            ),
        ),
        indirect=["endpoint", "exception"],
    )
    def test_init(self, endpoint, name, wrapper, methods, expected_methods, exception):
        with exception:
            route = Route("/", endpoint, methods=methods, include_in_schema=False)

            assert route.path == "/"
            assert isinstance(route.app, wrapper)
            assert route.endpoint == endpoint
            assert route.name == name
            assert route.include_in_schema is False
            assert route.methods == expected_methods

    @pytest.mark.parametrize(
        ["scope_type", "handle_call"],
        (
            pytest.param("http", True, id="http"),
            pytest.param("wrong", False, id="wrong"),
        ),
    )
    async def test_call(self, route, asgi_scope, asgi_receive, asgi_send, scope_type, handle_call):
        scope = types.Scope({**asgi_scope, "type": scope_type})
        route_scope = types.Scope({"foo": "bar"})
        handle = AsyncMock()
        expected_calls = [call(types.Scope({**scope, **route_scope}), asgi_receive, asgi_send)] if handle_call else []

        with patch.object(route, "handle", new=handle), patch.object(route, "route_scope", return_value=route_scope):
            await route(scope, asgi_receive, asgi_send)

        assert handle.call_args_list == expected_calls

    def test_eq(self):
        def foo(): ...

        assert Route("/", foo, methods={"GET"}) == Route("/", foo, methods={"GET"})
        assert Route("/", foo, methods={"GET"}) != Route("/", foo, methods={"POST"})

    def test_repr(self):
        def foo(): ...

        assert repr(Route("/", foo)) == "Route(path='/', name='foo', methods=['GET', 'HEAD'])"

    @pytest.mark.parametrize(
        ["endpoint", "result"],
        (
            pytest.param("function", False, id="function"),
            pytest.param("endpoint", True, id="endpoint"),
        ),
        indirect=["endpoint"],
    )
    def test_is_endpoint(self, endpoint, result):
        assert Route.is_endpoint(endpoint) is result

    @pytest.mark.parametrize(
        ["endpoint", "methods", "expected_methods"],
        (
            pytest.param("function", {"GET"}, {"GET", "HEAD"}, id="function_explicit_methods"),
            pytest.param("function", None, {"GET", "HEAD"}, id="function_no_methods"),
            pytest.param("function", {"POST"}, {"POST"}, id="function_no_get_method"),
            pytest.param("endpoint", {"GET"}, {"GET", "HEAD"}, id="endpoint_explicit_methods"),
            pytest.param("endpoint", None, {"GET", "HEAD", "POST"}, id="endpoint_no_methods"),
            pytest.param("function", {"POST"}, {"POST"}, id="endpoint_no_get_method"),
        ),
        indirect=["endpoint"],
    )
    def test_endpoint_handlers(self, endpoint, methods, expected_methods):
        result = {m: getattr(endpoint, m.lower() if m != "HEAD" else "get", endpoint) for m in expected_methods}

        route = Route("/", endpoint, methods=methods)

        assert route.endpoint_handlers() == result

    @pytest.mark.parametrize(
        ["scope_type", "scope_method", "path_match_return", "result"],
        (
            pytest.param("http", "GET", BaseRoute.Match.full, BaseRoute.Match.full, id="match"),
            pytest.param("http", "POST", BaseRoute.Match.full, BaseRoute.Match.partial, id="partial"),
            pytest.param("http", "GET", BaseRoute.Match.none, BaseRoute.Match.none, id="no_match"),
            pytest.param("websocket", "GET", None, BaseRoute.Match.none, id="wrong_scope_type"),
        ),
    )
    def test_match(self, scope_type, scope_method, path_match_return, result, asgi_scope):
        def foo(): ...

        route = Route("/", foo, methods={"GET"})

        asgi_scope["type"] = scope_type
        asgi_scope["method"] = scope_method

        with patch.object(BaseRoute, "match", return_value=path_match_return):
            assert route.match(asgi_scope) == result



import typing as t

from flama import types

if t.TYPE_CHECKING:
    from flama import http

HandlerException = t.TypeVar("HandlerException", bound=Exception)
Handler = t.Callable[
    [types.Scope, types.Receive, types.Send, HandlerException],
    t.Union[t.Optional["http.Response"], t.Awaitable[t.Optional["http.Response"]]],
]



import typing as t

from flama import compat

__all__ = [
    "OpenAPISpecInfoContact",
    "OpenAPISpecInfoLicense",
    "OpenAPISpecInfo",
    "OpenAPISpecServerVariable",
    "OpenAPISpecServer",
    "OpenAPISpecExternalDocs",
    "OpenAPISpecTag",
    "OpenAPISpecSecurity",
    "OpenAPISpec",
]


class OpenAPISpecInfoContact(t.TypedDict):
    name: str
    url: str
    email: str


class OpenAPISpecInfoLicense(t.TypedDict):
    name: str
    identifier: compat.NotRequired[t.Optional[str]]
    url: compat.NotRequired[t.Optional[str]]


class OpenAPISpecInfo(t.TypedDict):
    title: str
    summary: compat.NotRequired[t.Optional[str]]
    description: compat.NotRequired[t.Optional[str]]
    termsOfService: compat.NotRequired[t.Optional[str]]
    contact: compat.NotRequired[t.Optional[OpenAPISpecInfoContact]]
    license: compat.NotRequired[t.Optional[OpenAPISpecInfoLicense]]
    version: str


class OpenAPISpecServerVariable(t.TypedDict):
    default: str
    enum: compat.NotRequired[t.Optional[list[str]]]
    description: compat.NotRequired[t.Optional[str]]


class OpenAPISpecServer(t.TypedDict):
    url: str
    description: compat.NotRequired[t.Optional[str]]
    variables: compat.NotRequired[t.Optional[dict[str, OpenAPISpecServerVariable]]]


class OpenAPISpecExternalDocs(t.TypedDict):
    url: str
    description: compat.NotRequired[t.Optional[str]]


class OpenAPISpecTag(t.TypedDict):
    name: str
    description: compat.NotRequired[t.Optional[str]]
    externalDocs: compat.NotRequired[t.Optional[OpenAPISpecExternalDocs]]


OpenAPISpecSecurity = dict[str, list[str]]


class OpenAPISpec(t.TypedDict):
    info: OpenAPISpecInfo
    servers: compat.NotRequired[t.Optional[list[OpenAPISpecServer]]]
    security: compat.NotRequired[t.Optional[list[OpenAPISpecSecurity]]]
    tags: compat.NotRequired[t.Optional[list[OpenAPISpecTag]]]
    externalDocs: compat.NotRequired[t.Optional[OpenAPISpecExternalDocs]]



import abc
import logging
import re
import typing as t

from flama import Flama, concurrency, exceptions, types
from flama.telemetry.data_structures import Error, Response, TelemetryData

logger = logging.getLogger(__name__)

__all__ = ["TelemetryMiddleware"]


PROJECT = "vortico-core"
SERVICE = "elektrococo"
TOPIC_ID = "telemetry-bus"

HookFunction = t.Callable[[TelemetryData], t.Union[None, t.Awaitable[None]]]


class Wrapper(abc.ABC):
    def __init__(self, app: Flama, data: TelemetryData) -> None:
        self.app = app
        self.data = data

    @classmethod
    def build(cls, type: t.Literal["http", "websocket"], app: Flama, data: TelemetryData) -> "Wrapper":
        if type == "websocket":
            return WebSocketWrapper(app, data)

        return HTTPWrapper(app, data)

    async def __call__(self, scope: types.Scope, receive: types.Receive, send: types.Send) -> None:
        self._scope = scope
        self._receive = receive
        self._send = send
        self._response_body = b""
        self._response_headers = None
        self._response_status_code = None

        try:
            await self.app(self._scope, self.receive, self.send)
            self.data.response = Response(
                headers=self._response_headers, body=self._response_body, status_code=self._response_status_code
            )
        except Exception as e:
            self.data.error = await Error.from_exception(exception=e)
            raise

    @abc.abstractmethod
    async def receive(self) -> types.Message: ...

    @abc.abstractmethod
    async def send(self, message: types.Message) -> None: ...


class HTTPWrapper(Wrapper):
    async def receive(self) -> types.Message:
        message = await self._receive()

        if message["type"] == "http.request":
            self.data.request.body += message.get("body", b"")

        return message

    async def send(self, message: types.Message) -> None:
        if message["type"] == "http.response.start":
            self._response_headers = {k.decode(): v.decode() for (k, v) in message.get("headers", [])}
            self._response_status_code = message.get("status")
        elif message["type"] == "http.response.body":
            self._response_body += message.get("body", b"")

        await self._send(message)


class WebSocketWrapper(Wrapper):
    async def receive(self) -> types.Message:
        message = await self._receive()

        if message["type"] == "websocket.receive":
            self._response_body += message.get("body", b"")
        elif message["type"] == "websocket.disconnect":
            self._response_status_code = message.get("code", None)
            self._response_body = message.get("reason", "").encode()

        return message

    async def send(self, message: types.Message) -> None:
        if message["type"] == "websocket.send":
            self.data.request.body += message.get("bytes", message.get("text", "").encode())
        elif message["type"] == "websocket.close":
            self._response_status_code = message.get("code")
            self._response_body = message.get("reason", "").encode()

        await self._send(message)


class TelemetryDataCollector:
    data: TelemetryData

    def __init__(self, app: Flama, scope: types.Scope, receive: types.Receive, send: types.Send) -> None:
        self.app = app
        self._scope = scope
        self._receive = receive
        self._send = send

    @classmethod
    async def build(
        cls, app: Flama, scope: types.Scope, receive: types.Receive, send: types.Send
    ) -> "TelemetryDataCollector":
        self = cls(app, scope, receive, send)
        self.data = await TelemetryData.from_scope(scope=scope, receive=receive, send=send)
        return self

    async def __call__(self) -> None:
        await Wrapper.build(self._scope["type"], self.app, self.data)(
            scope=self._scope, receive=self._receive, send=self._send
        )


class TelemetryMiddleware:
    def __init__(
        self,
        app: types.App,
        *,
        log_level: int = logging.NOTSET,
        before: t.Optional[HookFunction] = None,
        after: t.Optional[HookFunction] = None,
        tag: str = "telemetry",
        ignored: list[str] = [],
    ) -> None:
        self.app: Flama = t.cast(Flama, app)
        self._log_level = log_level
        self._before = before
        self._after = after
        self._tag = tag
        self._ignored = [re.compile(x) for x in ignored]

    async def before(self, data: TelemetryData):
        if self._before:
            await concurrency.run(self._before, data)

    async def after(self, data: TelemetryData):
        if self._after:
            await concurrency.run(self._after, data)

    def _get_tag(self, scope: "types.Scope") -> bool:
        try:
            app: Flama = scope["app"]
            route, _ = app.router.resolve_route(scope)
            return route.tags.get(self._tag, True)
        except (exceptions.MethodNotAllowedException, exceptions.NotFoundException):
            return False

    async def __call__(self, scope: types.Scope, receive: types.Receive, send: types.Send) -> None:
        if (
            scope["type"] not in ("http", "websocket")
            or any(pattern.match(scope["path"]) for pattern in self._ignored)
            or not self._get_tag(scope)
        ):
            await self.app(scope, receive, send)
            return

        collector = await TelemetryDataCollector.build(self.app, scope, receive, send)

        await self.before(collector.data)

        try:
            await collector()
        finally:
            await self.after(collector.data)
            logger.log(self._log_level, "Telemetry: %s", str(collector.data))



import typing as t

import pytest

from flama.injection import ComponentNotFound
from flama.injection.components import Component, Components
from flama.injection.resolver import (
    EMPTY,
    ComponentNode,
    ContextNode,
    Parameter,
    ParameterNode,
    ResolutionTree,
    Resolver,
)

CustomInt = t.NewType("CustomInt", int)
Foo = t.NewType("Foo", int)
Bar = t.NewType("Bar", int)
Wrong = t.NewType("Wrong", int)
Unhandled = t.NewType("Unhandled", int)


class FooComponent(Component):
    def resolve(self, x: int) -> Foo:
        return Foo(x)


foo_component = FooComponent()


class BarComponent(Component):
    def resolve(self, parameter: Parameter, data: dict) -> Bar:
        return Bar(data[parameter.name])


bar_component = BarComponent()


class WrongComponent(Component):
    def resolve(self, x: Unhandled) -> Wrong:
        return Wrong(0)


wrong_component = WrongComponent()


class TestCaseResolutionTree:
    @pytest.fixture(scope="function")
    def context_types(self):
        return {CustomInt: "y"}

    @pytest.fixture(scope="function")
    def components(self):
        return Components([foo_component, bar_component, wrong_component])

    @pytest.mark.parametrize(
        ["parameter", "expected_tree", "expected_context", "expected_parameters", "expected_components", "exception"],
        (
            pytest.param(
                Parameter("x", int, EMPTY),
                ResolutionTree(root=ContextNode(name="x", parameter=Parameter(name="x", annotation=int), nodes=[])),
                [("x", Parameter(name="x", annotation=int, default=EMPTY))],
                [],
                [],
                None,
                id="context_builtin_type",
            ),
            pytest.param(
                Parameter("y", CustomInt, EMPTY),
                ResolutionTree(
                    root=ContextNode(name="y", parameter=Parameter(name="y", annotation=CustomInt), nodes=[])
                ),
                [("y", Parameter(name="y", annotation=CustomInt, default=EMPTY))],
                [],
                [],
                None,
                id="context_custom_type",
            ),
            pytest.param(
                Parameter("foo", Foo, EMPTY),
                ResolutionTree(
                    root=ComponentNode(
                        name="foo",
                        parameter=Parameter(name="foo", annotation=Foo),
                        nodes=[ContextNode(name="x", parameter=Parameter(name="x", annotation=int), nodes=[])],
                        component=foo_component,
                    )
                ),
                [("x", Parameter(name="x", annotation=int, default=EMPTY))],
                [],
                [("foo", foo_component)],
                None,
                id="component",
            ),
            pytest.param(
                Parameter("bar", Bar, EMPTY),
                ResolutionTree(
                    root=ComponentNode(
                        name="bar",
                        parameter=Parameter(name="bar", annotation=Bar),
                        nodes=[
                            ParameterNode(name="parameter", parameter=Parameter(name="bar", annotation=Bar), nodes=[]),
                            ContextNode(name="data", parameter=Parameter(name="data", annotation=dict), nodes=[]),
                        ],
                        component=bar_component,
                    )
                ),
                [("data", Parameter(name="data", annotation=dict))],
                [Parameter(name="bar", annotation=Bar)],
                [("bar", bar_component)],
                None,
                id="component_using_its_parameter",
            ),
            pytest.param(
                Parameter("wrong", Unhandled, EMPTY),
                None,
                None,
                None,
                None,
                ComponentNotFound(parameter=Parameter(name="wrong", annotation=Unhandled)),
                id="not_found",
            ),
            pytest.param(
                Parameter("wrong", Wrong, EMPTY),
                None,
                None,
                None,
                None,
                ComponentNotFound(parameter=Parameter(name="x", annotation=Unhandled), component=wrong_component),
                id="not_found_in_component",
            ),
        ),
        indirect=["exception"],
    )
    def test_build(
        self,
        parameter,
        expected_tree,
        expected_context,
        expected_parameters,
        expected_components,
        exception,
        context_types,
        components,
    ):
        with exception:
            tree = ResolutionTree.build(parameter, context_types, components)

            assert tree == expected_tree

            assert tree.context == expected_context
            assert tree.parameters == expected_parameters
            assert tree.components == expected_components

    @pytest.mark.parametrize(
        ["parameter", "context", "expected_value"],
        (
            pytest.param(
                Parameter("x", int, EMPTY),
                {"x": 1},
                1,
                id="context_builtin_type",
            ),
            pytest.param(
                Parameter("y", CustomInt, EMPTY),
                {"y": 1},
                1,
                id="context_custom_type",
            ),
            pytest.param(
                Parameter("foo", Foo, EMPTY),
                {"x": 1},
                Foo(1),
                id="component",
            ),
            pytest.param(
                Parameter("bar", Bar, EMPTY),
                {"data": {"bar": 2}},
                Bar(2),
                id="component_using_its_parameter",
            ),
        ),
    )
    async def test_context(self, parameter, context, expected_value, context_types, components):
        tree = ResolutionTree.build(parameter, context_types, components)

        value = await tree.value(context)

        assert value == expected_value


class TestCaseResolver:
    @pytest.fixture
    def resolver(self):
        return Resolver({"y": CustomInt}, Components([foo_component, bar_component, wrong_component]))

    @pytest.mark.parametrize(
        ["parameter", "expected_tree", "cached", "exception"],
        (
            pytest.param(
                Parameter("_root", int),
                ResolutionTree(
                    root=ContextNode(name="_root", parameter=Parameter(name="_root", annotation=int), nodes=[])
                ),
                False,
                None,
                id="context_builtin_type",
            ),
            pytest.param(
                Parameter("_root", CustomInt),
                ResolutionTree(
                    root=ContextNode(name="_root", parameter=Parameter(name="y", annotation=CustomInt), nodes=[])
                ),
                False,
                None,
                id="context_custom_type",
            ),
            pytest.param(
                Parameter("_root", Foo),
                ResolutionTree(
                    root=ComponentNode(
                        name="_root",
                        parameter=Parameter(name="_root", annotation=Foo),
                        nodes=[
                            ContextNode(name="x", parameter=Parameter(name="x", annotation=int), nodes=[]),
                        ],
                        component=foo_component,
                    )
                ),
                True,
                None,
                id="component",
            ),
            pytest.param(
                Parameter("_root", Bar),
                ResolutionTree(
                    root=ComponentNode(
                        name="_root",
                        parameter=Parameter(name="_root", annotation=Bar),
                        nodes=[
                            ParameterNode(
                                name="parameter", parameter=Parameter(name="_root", annotation=Bar), nodes=[]
                            ),
                            ContextNode(name="data", parameter=Parameter(name="data", annotation=dict), nodes=[]),
                        ],
                        component=bar_component,
                    )
                ),
                False,
                None,
                id="component_using_its_parameter",
            ),
            pytest.param(
                Parameter("_root", Unhandled),
                None,
                None,
                (ComponentNotFound, "No component able to handle parameter '_root'"),
                id="not_found",
            ),
            pytest.param(
                Parameter("_root", Wrong),
                None,
                None,
                (ComponentNotFound, "No component able to handle parameter 'x' in component 'WrongComponent'"),
                id="not_found_in_component",
            ),
        ),
        indirect=["exception"],
    )
    def test_resolve(self, parameter, expected_tree, cached, exception, resolver):
        with exception:
            assert resolver._cache == {}
            assert resolver.resolve(parameter) == expected_tree

            if cached:
                assert hash(parameter.annotation) in resolver._cache
                assert resolver.resolve(parameter) == expected_tree



import asyncio
import uuid
from unittest.mock import AsyncMock, MagicMock, call, patch

import pytest

from flama import Component, Flama, Module, exceptions, http, routing, types, websockets
from flama.ddd.components import WorkerComponent
from flama.events import Events
from flama.injection.injector import Injector
from flama.middleware import Middleware, MiddlewareStack
from flama.models import ModelsModule
from flama.resources import Resource, ResourcesModule, resource_method
from flama.schemas.modules import SchemaModule
from flama.types.applications import AppStatus
from flama.url import URL

DEFAULT_MODULES = [ResourcesModule, SchemaModule, ModelsModule]


class TestCaseFlama:
    @pytest.fixture(scope="function")
    def component(self):
        return MagicMock(spec=Component)

    @pytest.fixture(scope="function")
    def module(self):
        class Foo(Module):
            name = "foo"

        return Foo

    @pytest.fixture(scope="function")
    def middleware(self):
        return MagicMock(spec=Middleware)

    @pytest.fixture
    def app(self):
        return Flama(schema=None, docs=None)

    @pytest.fixture(scope="function")
    def tags(self):
        return {"tag": "foo", "list_tag": ["foo", "bar"], "dict_tag": {"foo": "bar"}}

    def test_init(self, module, component):
        component_obj = component()

        app = Flama(components=[component_obj], modules={module()})

        # Check router and main app
        assert isinstance(app.app, routing.Router)
        assert isinstance(app.router, routing.Router)
        # Check injector
        assert isinstance(app._injector, Injector)
        assert app._injector.context_types == {
            "scope": types.Scope,
            "receive": types.Receive,
            "send": types.Send,
            "exc": Exception,
            "app": Flama,
            "route": routing.BaseRoute,
            "request": http.Request,
            "response": http.Response,
            "websocket": websockets.WebSocket,
            "websocket_message": types.Message,
            "websocket_encoding": types.Encoding,
            "websocket_code": types.Code,
        }
        assert app._injector.components == []
        # Check middleware
        assert isinstance(app.middleware, MiddlewareStack)
        assert app.middleware
        # Check modules
        assert app.modules == {*DEFAULT_MODULES, module}
        # Check components
        default_components = app.components[:1]
        assert isinstance(default_components[0], WorkerComponent)
        assert app.components == [*default_components, component_obj]
        # Check events
        assert app.events == Events(
            startup=[m.on_startup for m in app.modules.values()], shutdown=[m.on_shutdown for m in app.modules.values()]
        )

    def test_getattr(self, app):
        for name, module in app.modules.items():
            assert getattr(app, name) == module

        assert getattr(app, "wrong") is None

    @pytest.mark.parametrize(
        ["request_type", "app_status", "exception"],
        (
            pytest.param("http", types.AppStatus.READY, None, id="http_ready"),
            pytest.param(
                "http",
                types.AppStatus.NOT_STARTED,
                exceptions.ApplicationError("Application is not ready to process requests yet."),
                id="http_not_started",
            ),
            pytest.param(
                "http",
                types.AppStatus.SHUT_DOWN,
                exceptions.ApplicationError("Application is already shut down."),
                id="http_shut_down",
            ),
            pytest.param("websocket", types.AppStatus.READY, None, id="websocket_ready"),
            pytest.param(
                "websocket",
                types.AppStatus.NOT_STARTED,
                exceptions.ApplicationError("Application is not ready to process requests yet."),
                id="websocket_not_started",
            ),
            pytest.param(
                "websocket",
                types.AppStatus.SHUT_DOWN,
                exceptions.ApplicationError("Application is already shut down."),
                id="websocket_shut_down",
            ),
            pytest.param("lifespan", types.AppStatus.READY, None, id="lifespan_ready"),
            pytest.param("lifespan", types.AppStatus.NOT_STARTED, None, id="lifespan_not_started"),
            pytest.param("lifespan", types.AppStatus.SHUT_DOWN, None, id="lifespan_shut_down"),
        ),
        indirect=["exception"],
    )
    async def test_call(self, request_type, app_status, exception, app, asgi_scope, asgi_receive, asgi_send):
        asgi_scope["type"] = request_type
        app._status = app_status

        with exception, patch.object(app, "middleware", new=AsyncMock(spec=MiddlewareStack)):
            await app(asgi_scope, asgi_receive, asgi_send)

            assert "app" in asgi_scope
            assert asgi_scope["app"] == app
            assert "root_app" in asgi_scope
            assert asgi_scope["root_app"] == app
            assert app.middleware.call_args_list == [call(asgi_scope, asgi_receive, asgi_send)]

    def test_components(self, app):
        component = MagicMock(spec=Component)

        app.router.add_component(component)

        default_components = app.components[:1]
        assert app.components == [*default_components, component]

    def test_add_component(self, app, component):
        component_obj = component()

        with patch.object(app, "router", spec=routing.Router) as router_mock:
            app.add_component(component_obj)

        assert router_mock.add_component.call_args_list == [call(component_obj)]
        assert router_mock.build.call_args_list == [call(app)]

    def test_routes(self, app):
        expected_routes = [MagicMock()]
        with patch.object(app, "router", spec=routing.Router) as router_mock:
            router_mock.routes = expected_routes
            routes = app.routes

        assert routes == expected_routes

    def test_add_route(self, app, tags):
        def foo(): ...

        with patch.object(app, "router", spec=routing.Router) as router_mock:
            router_mock.add_route.return_value = foo
            route = app.add_route("/", foo, tags=tags)

        assert router_mock.add_route.call_args_list == [
            call(
                "/",
                foo,
                methods=None,
                name=None,
                include_in_schema=True,
                route=None,
                root=app,
                pagination=None,
                tags=tags,
            )
        ]
        assert route == foo

    def test_route(self, app, tags):
        with patch.object(app, "router", spec=routing.Router) as router_mock:

            @app.route("/", tags=tags)
            def foo(): ...

        assert router_mock.route.call_args_list == [
            call("/", methods=None, name=None, include_in_schema=True, root=app, pagination=None, tags=tags)
        ]

    def test_add_websocket_route(self, app, tags):
        def foo(): ...

        with patch.object(app, "router", spec=routing.Router) as router_mock:
            router_mock.add_websocket_route.return_value = foo
            route = app.add_websocket_route("/", foo, tags=tags)

        assert router_mock.add_websocket_route.call_args_list == [
            call("/", foo, name=None, route=None, root=app, pagination=None, tags=tags)
        ]
        assert route == foo

    def test_websocket_route(self, app, tags):
        with patch.object(app, "router", spec=routing.Router) as router_mock:

            @app.websocket_route("/", tags=tags)
            def foo(): ...

        assert router_mock.websocket_route.call_args_list == [
            call("/", name=None, root=app, pagination=None, tags=tags)
        ]

    def test_mount(self, app, tags):
        expected_mount = MagicMock()

        with patch.object(app, "router", spec=routing.Router) as router_mock:
            router_mock.mount.return_value = expected_mount
            mount = app.mount("/", expected_mount, tags=tags)

        assert router_mock.mount.call_args_list == [
            call("/", expected_mount, name=None, mount=None, root=app, tags=tags)
        ]
        assert mount == expected_mount

    def test_injector(self, app):
        assert isinstance(app.injector, Injector)

    def test_add_event_handler(self, app):
        handlers_before = app.events.startup.copy()

        def handler(): ...

        app.add_event_handler("startup", handler)

        assert app.events.startup == [*handlers_before, handler]

    def test_on_event(self, app):
        handlers_before = app.events.startup.copy()

        @app.on_event("startup")
        def handler(): ...

        assert app.events.startup == [*handlers_before, handler]

    @pytest.mark.parametrize(
        ["key", "handler"],
        (pytest.param(400, MagicMock(), id="status_code"), pytest.param(ValueError, MagicMock(), id="exception_class")),
    )
    def test_add_exception_handler(self, app, key, handler):
        expected_call = [call(key, handler)]

        with patch.object(app, "middleware", spec=MiddlewareStack):
            app.add_exception_handler(key, handler)
            assert app.middleware.add_exception_handler.call_args_list == expected_call

    def test_add_middleware(self, app):
        class FooMiddleware:
            def __init__(self, app: types.App) -> None: ...

            def __call__(self, scope: types.Scope, receive: types.Receive, send: types.Send): ...

        kwargs = {"foo": "bar"}

        with patch.object(app, "middleware", spec=MiddlewareStack):
            app.add_middleware(Middleware(FooMiddleware, **kwargs))
            assert len(app.middleware.add_middleware.call_args_list) == 1
            middleware = app.middleware.add_middleware.call_args[0][0]
            assert isinstance(middleware, Middleware)
            assert middleware.middleware == FooMiddleware
            assert middleware.kwargs == kwargs

    @pytest.mark.parametrize(
        ["resolve", "path_params", "resolution", "exception"],
        (
            pytest.param("foo", {}, URL(scheme="http", path="/foo"), None, id="plain"),
            pytest.param("foo", {"x": 1}, URL(scheme="http", path="/foo/1"), None, id="path_params"),
            pytest.param(
                "bar",
                {"x": uuid.UUID(int=1), "y": uuid.UUID(int=2)},
                URL(scheme="http", path=f"/bar/{uuid.UUID(int=1)}/y/{uuid.UUID(int=2)}"),
                None,
                id="multiple_path_params",
            ),
            pytest.param("foo:bar", {}, URL(scheme="http", path="/foo/bar"), None, id="nested"),
            pytest.param(
                "puppy:custom", {}, URL(scheme="http", path="/puppy/custom"), None, id="resource_custom_method"
            ),
            pytest.param(
                "not-found",
                {},
                None,
                exceptions.NotFoundException(name="not-found"),
                id="not_found",
            ),
        ),
        indirect=["exception"],
    )
    def test_resolve_url(self, app, resolve, path_params, resolution, exception):
        app.add_route(route=routing.Route("/foo", lambda: None, name="foo"))
        app.add_route(route=routing.Route("/foo/{x:int}", lambda: None, name="foo"))
        app.add_route(route=routing.Route("/bar/{x:uuid}/y/{y:uuid}", lambda: None, name="bar"))
        app.mount(mount=routing.Mount("/foo", routes=[routing.Route("/bar", lambda: None, name="bar")], name="foo"))

        @app.resources.resource("/puppy")
        class PuppyResource(Resource):
            name = "puppy"
            verbose_name = "Puppy"

            @resource_method("/custom")
            def custom(self): ...

        with exception:
            assert app.resolve_url(resolve, **path_params) == resolution

    def test_build_application(self, module):
        root_component = MagicMock(spec=Component)
        root_app = Flama(schema=None, docs=None, components=[root_component])
        root_app.add_get("/foo", lambda: {})

        assert len(root_app.router.routes) == 1
        root_default_components = root_app.components[:1]
        assert root_app.components == [*root_default_components, root_component]
        assert root_app.modules == DEFAULT_MODULES

        leaf_component = MagicMock(spec=Component)
        leaf_app = Flama(schema=None, docs=None, components=[leaf_component], modules={module()})
        leaf_app.add_get("/bar", lambda: {})

        assert len(leaf_app.router.routes) == 1
        leaf_default_components = leaf_app.components[:1]
        assert leaf_app.components == [*leaf_default_components, leaf_component]
        assert leaf_app.modules == [*DEFAULT_MODULES, module]

        root_app.mount("/app", app=leaf_app)

        assert len(root_app.router.routes) == 2
        # Check mount is initialized
        assert isinstance(root_app.routes[1], routing.Mount)
        mount_route = root_app.router.routes[1]
        assert mount_route.path == "/app"
        # Check router is created and initialized
        assert isinstance(mount_route.app, Flama)
        mount_app = mount_route.app
        assert isinstance(mount_app.app, routing.Router)
        mount_router = mount_app.app
        # Check components are collected across the entire tree
        assert root_app.components == [*root_default_components, root_component]
        assert mount_router.components == [*leaf_default_components, leaf_component, *root_app.components]
        # Check modules are isolated for each app
        assert mount_app.modules == [*DEFAULT_MODULES, module]
        assert root_app.modules == DEFAULT_MODULES

    def test_build_application_declarative(self, module):
        leaf_component = MagicMock(spec=Component)
        leaf_routes = [routing.Route("/bar", lambda: {})]
        leaf_app = Flama(routes=leaf_routes, schema=None, docs=None, components=[leaf_component], modules={module()})
        root_component = MagicMock(spec=Component)
        root_routes = [routing.Route("/foo", lambda: {}), routing.Mount("/app", app=leaf_app)]
        root_app = Flama(routes=root_routes, schema=None, docs=None, components=[root_component])

        assert len(root_app.router.routes) == 2
        # Check mount is initialized
        assert isinstance(root_app.routes[1], routing.Mount)
        mount_route = root_app.router.routes[1]
        assert mount_route.path == "/app"
        # Check router is created and initialized
        assert isinstance(mount_route.app, Flama)
        mount_app = mount_route.app
        assert isinstance(mount_app.app, routing.Router)
        mount_router = mount_app.app
        # Check components are collected across the entire tree
        root_default_components = root_app.components[:1]
        assert root_app.components == [*root_default_components, root_component]
        leaf_default_components = leaf_app.components[:1]
        assert mount_router.components == [*leaf_default_components, leaf_component, *root_app.components]
        # Check modules are isolated for each app
        assert mount_app.modules == [*DEFAULT_MODULES, module]
        assert root_app.modules == DEFAULT_MODULES

    async def test_call_complete_workflow(self):
        receive_queue = asyncio.Queue()
        send_queue = asyncio.Queue()

        async def receive() -> types.Message:
            return await receive_queue.get()

        async def send(message: types.Message) -> None:
            await send_queue.put(message)

        app = Flama(docs=None, schema=None)
        sub_app = Flama(docs=None, schema=None)
        sub_app.add_route("/bar/", lambda: {"foo": "bar"})
        app.mount("/foo", sub_app)

        assert app._status == AppStatus.NOT_STARTED
        assert sub_app._status == AppStatus.NOT_STARTED

        await receive_queue.put(types.Message({"type": "lifespan.startup"}))
        await app(types.Scope({"type": "lifespan"}), receive, send)
        assert await send_queue.get() == {"type": "lifespan.startup.complete"}

        assert app._status == AppStatus.READY
        assert sub_app._status == AppStatus.READY

        await app(types.Scope({"type": "http", "path": "/foo/bar/", "method": "GET"}), receive, send)
        assert await send_queue.get() == {
            "type": "http.response.start",
            "headers": [(b"content-length", b"13"), (b"content-type", b"application/json")],
            "status": 200,
        }
        assert await send_queue.get() == {"body": b'{"foo":"bar"}', "type": "http.response.body"}

        await receive_queue.put(types.Message({"type": "lifespan.shutdown"}))
        await app(types.Scope({"type": "lifespan"}), receive, send)
        assert await send_queue.get() == {"type": "lifespan.shutdown.complete"}

        assert app._status == AppStatus.SHUT_DOWN
        assert sub_app._status == AppStatus.SHUT_DOWN



import abc
import typing as t

if t.TYPE_CHECKING:
    from flama import http, types


class Codec(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    async def decode(self, item: t.Any, **options) -> t.Any: ...


class HTTPCodec(Codec):
    media_type: t.Optional[str] = None

    @abc.abstractmethod
    async def decode(self, item: "http.Request", **options) -> t.Any: ...


class WebsocketsCodec(Codec):
    encoding: t.Optional[str] = None

    @abc.abstractmethod
    async def decode(self, item: "types.Message", **options) -> t.Any: ...



import codecs
import importlib.metadata
import logging
import math
import pickle
import typing as t
import warnings

from flama import exceptions
from flama.serialize import types
from flama.serialize.base import Serializer

if t.TYPE_CHECKING:
    from flama.types import JSONField, JSONSchema

logger = logging.getLogger(__name__)


class SKLearnSerializer(Serializer):
    lib = types.Framework.sklearn

    def dump(self, obj: t.Any, **kwargs) -> bytes:
        return codecs.encode(pickle.dumps(obj), "base64")

    def load(self, model: bytes, **kwargs) -> t.Any:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            model = pickle.loads(codecs.decode(model, "base64"))

        return model

    def _info(self, data) -> "JSONField":
        if isinstance(data, (int, bool, str)):
            return data

        if isinstance(data, float):
            return None if math.isnan(data) else data

        if isinstance(data, dict):
            return {k: self._info(v) for k, v in data.items()}

        if isinstance(data, (list, tuple, set)):
            return [self._info(i) for i in data]

        return None

    def info(self, model: t.Any) -> t.Optional["JSONSchema"]:
        try:
            return self._info(model.get_params())  # type: ignore
        except:  # noqa
            logger.exception("Cannot collect info from model")
            return None

    def version(self) -> str:
        try:
            return importlib.metadata.version("scikit-learn")
        except Exception:  # noqa
            raise exceptions.FrameworkNotInstalled("scikit-learn")



import datetime
import http
import importlib.metadata
import uuid
from unittest.mock import AsyncMock, MagicMock, call, patch

import pytest

from flama import Flama, authentication, types
from flama.middleware import Middleware
from flama.telemetry import Authentication, Endpoint, Error, Request, Response, TelemetryData, TelemetryMiddleware

SECRET = uuid.UUID(int=0)

TOKEN = (
    "eyJhbGciOiAiSFMyNTYiLCAidHlwIjogIkpXVCJ9.eyJkYXRhIjogeyJmb28iOiAiYmFyIn0sICJpYXQiOiAwfQ==.J3zdedMZSFNOimstjJat0V"
    "28rM_b1UU62XCp9dg_5kg="
)
DECODED_TOKEN = authentication.JWT.decode(TOKEN.encode(), SECRET.bytes)


class TestCaseTelemetryMiddleware:
    # TODO: WebSocketWrapper is not tested

    @pytest.fixture(scope="function")
    def app(self):
        return Flama(
            schema=None,
            docs=None,
            components=[
                authentication.AccessTokenComponent(secret=SECRET.bytes),
                authentication.RefreshTokenComponent(secret=SECRET.bytes),
            ],
        )

    @pytest.fixture(scope="function", autouse=True)
    def add_endpoints(self, app):
        @app.post("/{x:int}/", name="foo", tags={"foo": "bar"})
        def root(x: int, y: int, body: types.Body):
            return {"x": x, "y": y, "body": body}

        @app.post("/error/", name="error")
        def error():
            raise ValueError("foo")

        @app.post("/ignored/", name="ignored")
        def ignored():
            return "ignored"

        @app.post("/explicit-off/", name="explicit-off", tags={"telemetry": False})
        def explicit_off():
            return "explicit_off"

    @pytest.mark.parametrize(
        [
            "path",
            "request_params",
            "request_body",
            "request_cookies",
            "status_code",
            "response",
            "exception",
            "before",
            "after",
            "data",
        ],
        [
            pytest.param(
                "/1/",
                {"y": 1},
                b"body",
                {"access_token": TOKEN},
                http.HTTPStatus.OK,
                {"x": 1, "y": 1, "body": "body"},
                None,
                None,
                None,
                None,
                id="ok_no_hooks",
            ),
            pytest.param(
                "/1/",
                {"y": 1},
                b"body",
                {"access_token": TOKEN},
                http.HTTPStatus.OK,
                {"x": 1, "y": 1, "body": "body"},
                None,
                MagicMock(),
                MagicMock(),
                TelemetryData(
                    type="http",
                    endpoint=Endpoint(path="/{x}/", name="foo", tags={"foo": "bar"}),
                    authentication=Authentication(access=authentication.AccessToken(DECODED_TOKEN), refresh=None),
                    request=Request(
                        headers={
                            "host": "localapp",
                            "accept": "*/*",
                            "accept-encoding": "gzip, deflate",
                            "connection": "keep-alive",
                            "user-agent": f"flama/{importlib.metadata.version('flama')}",
                            "content-length": "4",
                            "cookie": "access_token=eyJhbGciOiAiSFMyNTYiLCAidHlwIjogIkpXVCJ9.eyJkYXRhIjogeyJmb28iOiAiY"
                            "mFyIn0sICJpYXQiOiAwfQ==.J3zdedMZSFNOimstjJat0V28rM_b1UU62XCp9dg_5kg=",
                        },
                        cookies={
                            "access_token": {
                                "expires": "",
                                "path": "",
                                "comment": "",
                                "domain": "",
                                "max-age": "",
                                "secure": "",
                                "httponly": "",
                                "version": "",
                                "samesite": "",
                                "value": "eyJhbGciOiAiSFMyNTYiLCAidHlwIjogIkpXVCJ9.eyJkYXRhIjogeyJmb28iOiAiYmFyIn0sICJ"
                                "pYXQiOiAwfQ==.J3zdedMZSFNOimstjJat0V28rM_b1UU62XCp9dg_5kg=",
                            }
                        },
                        query_parameters={"y": "1"},
                        path_parameters={"x": 1},
                        body=b"body",
                    ),
                    response=Response(
                        headers={"content-length": "27", "content-type": "application/json"},
                        body=b'{"x":1,"y":1,"body":"body"}',
                        status_code=http.HTTPStatus.OK,
                    ),
                ),
                id="ok_sync_hooks",
            ),
            pytest.param(
                "/1/",
                {"y": 1},
                b"body",
                {"access_token": TOKEN},
                http.HTTPStatus.OK,
                {"x": 1, "y": 1, "body": "body"},
                None,
                AsyncMock(),
                AsyncMock(),
                TelemetryData(
                    type="http",
                    endpoint=Endpoint(path="/{x}/", name="foo", tags={"foo": "bar"}),
                    authentication=Authentication(access=authentication.AccessToken(DECODED_TOKEN), refresh=None),
                    request=Request(
                        headers={
                            "host": "localapp",
                            "accept": "*/*",
                            "accept-encoding": "gzip, deflate",
                            "connection": "keep-alive",
                            "user-agent": f"flama/{importlib.metadata.version('flama')}",
                            "content-length": "4",
                            "cookie": "access_token=eyJhbGciOiAiSFMyNTYiLCAidHlwIjogIkpXVCJ9.eyJkYXRhIjogeyJmb28iOiAiY"
                            "mFyIn0sICJpYXQiOiAwfQ==.J3zdedMZSFNOimstjJat0V28rM_b1UU62XCp9dg_5kg=",
                        },
                        cookies={
                            "access_token": {
                                "expires": "",
                                "path": "",
                                "comment": "",
                                "domain": "",
                                "max-age": "",
                                "secure": "",
                                "httponly": "",
                                "version": "",
                                "samesite": "",
                                "value": "eyJhbGciOiAiSFMyNTYiLCAidHlwIjogIkpXVCJ9.eyJkYXRhIjogeyJmb28iOiAiYmFyIn0sICJ"
                                "pYXQiOiAwfQ==.J3zdedMZSFNOimstjJat0V28rM_b1UU62XCp9dg_5kg=",
                            }
                        },
                        query_parameters={"y": "1"},
                        path_parameters={"x": 1},
                        body=b"body",
                    ),
                    response=Response(
                        headers={"content-length": "27", "content-type": "application/json"},
                        body=b'{"x":1,"y":1,"body":"body"}',
                        status_code=http.HTTPStatus.OK,
                    ),
                ),
                id="ok_async_hooks",
            ),
            pytest.param(
                "/error/",
                {},
                None,
                {},
                http.HTTPStatus.OK,
                None,
                ValueError("foo"),
                None,
                None,
                None,
                id="error_no_hooks",
            ),
            pytest.param(
                "/error/",
                {},
                None,
                {},
                http.HTTPStatus.OK,
                None,
                ValueError("foo"),
                MagicMock(),
                MagicMock(),
                TelemetryData(
                    type="http",
                    endpoint=Endpoint(path="/error/", name="error", tags={}),
                    authentication=Authentication(access=None, refresh=None),
                    request=Request(
                        headers={
                            "host": "localapp",
                            "accept": "*/*",
                            "accept-encoding": "gzip, deflate",
                            "connection": "keep-alive",
                            "user-agent": f"flama/{importlib.metadata.version('flama')}",
                            "content-length": "0",
                        },
                        cookies={},
                        query_parameters={},
                        path_parameters={},
                        body=b"",
                    ),
                    error=Error(detail="foo", status_code=None),
                ),
                id="error_sync_hooks",
            ),
            pytest.param(
                "/error/",
                {},
                None,
                {},
                http.HTTPStatus.OK,
                None,
                ValueError("foo"),
                AsyncMock(),
                AsyncMock(),
                TelemetryData(
                    type="http",
                    endpoint=Endpoint(path="/error/", name="error", tags={}),
                    authentication=Authentication(access=None, refresh=None),
                    request=Request(
                        headers={
                            "host": "localapp",
                            "accept": "*/*",
                            "accept-encoding": "gzip, deflate",
                            "connection": "keep-alive",
                            "user-agent": f"flama/{importlib.metadata.version('flama')}",
                            "content-length": "0",
                        },
                        cookies={},
                        query_parameters={},
                        path_parameters={},
                        body=b"",
                    ),
                    error=Error(detail="foo", status_code=None),
                ),
                id="error_async_hooks",
            ),
            pytest.param(
                "/ignored/",
                {},
                None,
                {},
                http.HTTPStatus.OK,
                "ignored",
                None,
                AsyncMock(),
                AsyncMock(),
                None,
                id="ignored",
            ),
            pytest.param(
                "/explicit-off/",
                {},
                None,
                {},
                http.HTTPStatus.OK,
                "explicit_off",
                None,
                AsyncMock(),
                AsyncMock(),
                None,
                id="explicit_off",
            ),
            pytest.param(
                "/not-found/",
                {},
                None,
                {},
                http.HTTPStatus.NOT_FOUND,
                {"status_code": 404, "detail": "Not Found", "error": "HTTPException"},
                None,
                None,
                None,
                None,
                id="not_found",
            ),
        ],
        indirect=["exception"],
    )
    async def test_request(
        self,
        app,
        client,
        path,
        request_params,
        request_body,
        request_cookies,
        status_code,
        response,
        exception,
        before,
        after,
        data,
    ):
        app.add_middleware(Middleware(TelemetryMiddleware, before=before, after=after, ignored=[r"/ignored.*"]))

        client.cookies = request_cookies

        now = datetime.datetime.now()

        if data:
            data.request.timestamp = now
            if data.response:
                data.response.timestamp = now
            if data.error:
                data.error.timestamp = now

        with exception, patch("datetime.datetime", MagicMock(now=MagicMock(return_value=now))):
            r = await client.post(path, params=request_params, content=request_body)

            assert r.status_code == status_code
            assert r.json() == response

        if before:
            assert before.call_args_list == ([call(data)] if data else [])

        if after:
            assert after.call_args_list == ([call(data)] if data else [])



import typing as t

__all__ = [
    "Empty",
    "RepositoryException",
    "IntegrityError",
    "NotFoundError",
    "AlreadyExistsError",
    "MultipleRecordsError",
]


class Empty(Exception): ...


class RepositoryException(Exception): ...


class ResourceException(RepositoryException):
    _error_message: t.ClassVar[str] = "exception"

    def __init__(self, *, resource: str, id: t.Any = None, detail: str = "") -> None:
        super().__init__()
        self.resource = resource
        self.id = id
        self.detail = detail

    def __str__(self) -> str:
        return (
            f"Resource '{self.resource}'"
            + (f" ({self.id})" if self.id else "")
            + f" {self._error_message}"
            + (f" ({self.detail})" if self.detail else "")
        )

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}("{self.__str__()}")'


class IntegrityError(ResourceException):
    _error_message = "integrity failed"


class NotFoundError(ResourceException):
    _error_message = "not found"


class AlreadyExistsError(ResourceException):
    _error_message = "already exists"


class MultipleRecordsError(ResourceException):
    _error_message = "multiple records found"



import abc
import typing as t

from flama.serialize.types import Framework

if t.TYPE_CHECKING:
    from flama.types import JSONSchema

__all__ = ["Serializer"]


class Serializer(metaclass=abc.ABCMeta):
    lib: t.ClassVar[Framework]

    @abc.abstractmethod
    def dump(self, obj: t.Any, **kwargs) -> bytes: ...

    @abc.abstractmethod
    def load(self, model: bytes, **kwargs) -> t.Any: ...

    @abc.abstractmethod
    def info(self, model: t.Any) -> t.Optional["JSONSchema"]: ...

    @abc.abstractmethod
    def version(self) -> str: ...



import pytest

from flama.ddd.workers import Worker


class TestCaseWorker:
    @pytest.fixture(scope="function")
    def worker(self, client):
        class FooWorker(Worker): ...

        return FooWorker(client.app)

    def test_init(self, app):
        worker = Worker(app)

        assert worker._app == app



import pytest

from flama.applications import Flama
from flama.models import ModelResource, ModelResourceType
from flama.resources.routing import ResourceRoute


class TestCaseResourcesModule:
    @pytest.fixture(scope="function")
    def app(self):
        return Flama(schema=None, docs=None)

    @pytest.fixture(scope="function")
    def tags(self):
        return {"inspect": {"tag": "inspect"}, "predict": {"tag": "predict"}}

    def test_add_model(self, app, model, component, tags):
        component_ = component

        class PuppyModelResource(ModelResource, metaclass=ModelResourceType):
            name = "puppy"
            verbose_name = "Puppy"
            component = component_

        resource = app.models.add_model_resource("/", PuppyModelResource, tags=tags)

        assert len(app.routes) == 1
        assert isinstance(app.routes[0], ResourceRoute)
        resource_route = app.routes[0]
        assert len(resource_route.routes) == 2
        assert [(route.path, route.methods, route.endpoint, route.tags) for route in resource_route.routes] == [
            ("/", {"HEAD", "GET"}, resource.inspect, {"tag": "inspect"}),
            ("/predict/", {"POST"}, resource.predict, {"tag": "predict"}),
        ]

    def test_add_model_decorator(self, app, model, component, tags):
        component_ = component

        class PuppyModelResource(ModelResource, metaclass=ModelResourceType):
            name = "puppy"
            verbose_name = "Puppy"
            component = component_

        resource = app.models.model_resource("/", tags=tags)(
            PuppyModelResource()
        )  # Apply decoration to an instance in order to check endpoints

        assert len(app.routes) == 1
        assert isinstance(app.routes[0], ResourceRoute)
        resource_route = app.routes[0]
        assert len(resource_route.routes) == 2
        assert [(route.path, route.methods, route.endpoint, route.tags) for route in resource_route.routes] == [
            ("/", {"HEAD", "GET"}, resource.inspect, {"tag": "inspect"}),
            ("/predict/", {"POST"}, resource.predict, {"tag": "predict"}),
        ]

    def test_mount_resource_declarative(self, model, component, tags):
        component_ = component

        class PuppyModelResource(ModelResource, metaclass=ModelResourceType):
            name = "puppy"
            verbose_name = "Puppy"
            component = component_

        route = ResourceRoute("/puppy/", PuppyModelResource, tags=tags)

        app = Flama(routes=[route], schema=None, docs=None)

        assert len(app.router.routes) == 1
        assert len(app.routes) == 1
        assert isinstance(app.routes[0], ResourceRoute)
        resource_route = app.routes[0]
        assert len(resource_route.routes) == 2
        assert [(route.path, route.methods, route.endpoint, route.tags) for route in resource_route.routes] == [
            ("/", {"HEAD", "GET"}, resource_route.resource.inspect, {"tag": "inspect"}),
            ("/predict/", {"POST"}, resource_route.resource.predict, {"tag": "predict"}),
        ]
        assert isinstance(resource_route.resource, PuppyModelResource)

    def test_add_model_resource(self, app, model, component, tags):
        component_ = component

        class PuppyModelResource(ModelResource, metaclass=ModelResourceType):
            name = "puppy"
            verbose_name = "Puppy"
            component = component_

        resource = PuppyModelResource()

        app.models.add_model_resource("/", resource, tags=tags)

        assert len(app.routes) == 1
        assert isinstance(app.routes[0], ResourceRoute)
        resource_route = app.routes[0]
        assert len(resource_route.routes) == 2
        assert [(route.path, route.methods, route.endpoint, route.tags) for route in resource_route.routes] == [
            ("/", {"HEAD", "GET"}, resource.inspect, {"tag": "inspect"}),
            ("/predict/", {"POST"}, resource.predict, {"tag": "predict"}),
        ]



from unittest.mock import Mock

import pytest

from flama import Flama
from flama.modules import Module, Modules


class TestCaseModule:
    @pytest.fixture
    def module(self):
        class FooModule(Module):
            name = "foo"

        return FooModule

    async def test_new_module(self):
        class FooModule(Module):
            name = "foo"

        module = FooModule()

        assert await module.on_startup() is None
        assert await module.on_shutdown() is None

    def test_new_module_no_name(self):
        with pytest.raises(AssertionError, match="Module 'FooModule' does not have a 'name' attribute."):

            class FooModule(Module): ...

    def test_init(self, module):
        m = module()

        assert not hasattr(m, "app")


class TestCaseModules:
    @pytest.fixture
    def app(self):
        return Mock(spec=Flama)

    @pytest.fixture
    def foo_module(self):
        class FooModule(Module):
            name = "foo"

        return FooModule()

    @pytest.fixture
    def bar_module(self):
        class BarModule(Module):
            name = "bar"

        return BarModule()

    @pytest.fixture
    def modules(self, app, foo_module, bar_module):
        return Modules(app, {foo_module, bar_module})

    def test_init(self, app, foo_module, bar_module):
        assert not hasattr(foo_module, "app")
        assert not hasattr(bar_module, "app")

        modules = Modules(app, {foo_module, bar_module})

        assert foo_module.app == app
        assert bar_module.app == app
        assert modules == {"foo": foo_module, "bar": bar_module}
        assert modules == [foo_module.__class__, bar_module.__class__]

    def test_init_collision(self, foo_module, app):
        class FooModule2(Module):
            name = "foo"

        with pytest.raises(AssertionError, match=r"Collision in module names: foo \(FooModule, FooModule2\)"):
            Modules(app, {foo_module, FooModule2()})



import dataclasses
import logging
import time
import typing as t

from flama.authentication import exceptions
from flama.authentication.jwt import claims
from flama.authentication.jwt.jws import JWS

logger = logging.getLogger(__name__)


__all__ = ["JWT"]

VALIDATORS = [
    claims.IssValidator,
    claims.SubValidator,
    claims.AudValidator,
    claims.ExpValidator,
    claims.NbfValidator,
    claims.IatValidator,
    claims.JtiValidator,
]


@dataclasses.dataclass(frozen=True)
class Header:
    """JWT header.

    It contains the metadata of the token. The header is represented as a dictionary, and it is
    validated when the token is decoded. The header must contain the algorithm used to sign the token.

    Additional information about the header can be found in the RFC 7519: https://tools.ietf.org/html/rfc7519
    """

    typ: str = "JWT"
    alg: t.Optional[str] = None
    cty: t.Optional[str] = None

    def to_dict(self) -> dict[str, t.Any]:
        """Return the header as a dictionary.

        The fields are sorted alphabetically and the None values are removed.

        :return: Header as a dictionary.
        """
        return dataclasses.asdict(
            self, dict_factory=lambda x: {k: v for k, v in sorted(x, key=lambda y: y[0]) if v is not None}
        )


@dataclasses.dataclass(frozen=True)
class Payload:
    """JWT payload.

    It contains the claims of the token. The claims are the statements about an entity (typically, the user) and
    additional data. The claims are represented as a dictionary, and they are validated when the token is decoded.

    Additional information about the claims can be found in the RFC 7519: https://tools.ietf.org/html/rfc7519

    The user data is stored in the `data` field, and it is encoded as a dictionary. This field is not part of the JWT
    standard and it is not validated when the token is decoded.
    """

    data: dict[str, t.Any]
    iss: t.Optional[str] = None
    sub: t.Optional[str] = None
    aud: t.Optional[str] = None
    exp: t.Optional[int] = None
    nbf: t.Optional[int] = None
    iat: t.Optional[int] = None
    jti: t.Optional[str] = None

    def __init__(
        self,
        data: t.Optional[dict[str, t.Any]] = None,
        iss: t.Optional[str] = None,
        sub: t.Optional[str] = None,
        aud: t.Optional[str] = None,
        exp: t.Optional[int] = None,
        nbf: t.Optional[int] = None,
        iat: t.Optional[int] = None,
        jti: t.Optional[str] = None,
        **kwargs: t.Any,
    ) -> None:
        """Initialize the payload.

        It contains the claims of the token. The claims are the statements about an entity (typically, the user) and
        additional data. The claims are represented as a dictionary, and they are validated when the token is decoded.

        Additional information about the claims can be found in the RFC 7519: https://tools.ietf.org/html/rfc7519

        The user data is stored in the `data` field, and it is encoded as a dictionary. This field is not part of the
        JWT standard and it is not validated when the token is decoded. The user data can be passed as a dictionary or
        as keyword arguments.

        :param data: User data.
        :param iss: Issuer.
        :param sub: Subject.
        :param aud: Audience.
        :param exp: Expiration time.
        :param nbf: Not before.
        :param iat: Issued at.
        :param jti: JWT ID.
        :param kwargs: User data.
        """
        object.__setattr__(self, "iss", iss)
        object.__setattr__(self, "sub", sub)
        object.__setattr__(self, "aud", aud)
        object.__setattr__(self, "exp", exp)
        object.__setattr__(self, "nbf", nbf)
        object.__setattr__(self, "iat", iat if iat is not None else int(time.time()))
        object.__setattr__(self, "jti", jti)
        object.__setattr__(self, "data", {**(data or {}), **kwargs})

    def to_dict(self) -> dict[str, t.Any]:
        """Return the payload as a dictionary.

        The fields are sorted alphabetically and the None values are removed.

        :return: Payload as a dictionary.
        """
        return dataclasses.asdict(
            self, dict_factory=lambda x: {k: v for k, v in sorted(x, key=lambda y: y[0]) if v is not None}
        )


@dataclasses.dataclass(frozen=True)
class JWT:
    """JSON Web Token (JWT) implementation.

    This is a convenient wrapper of the JWT methods from the authlib library. It is used to create and decode JWT
    tokens, and to validate the signature of the token.

    The token is signed using JSW, and the signature is validated using the algorithm specified in the header.
    """

    header: Header
    payload: Payload

    def __init__(self, header: dict[str, t.Any], payload: dict[str, t.Any]) -> None:
        object.__setattr__(self, "header", Header(**header))
        object.__setattr__(self, "payload", Payload(**payload))

    def encode(self, key: bytes) -> bytes:
        """Encode a JWT token.

        The token is signed using the given secret. The result is a JWT token with a format of:
        <header>.<payload>.<signature>

        :param key: Secret used to sign the token.
        :return: Encoded token.
        """
        return JWS.encode(header=self.header.to_dict(), payload=self.payload.to_dict(), key=key)

    @classmethod
    def decode(cls, token: bytes, key: bytes) -> "JWT":
        """Decode a JWT token.

        The token format must be: <header>.<payload>.<signature>

        :param token: Token to decode.
        :param key: Key used to sign the token.
        :return: An instance of JWT with the decoded token.
        :raises JWTDecodeException: If the token format is not correct.
        :raises JWTValidateException: If the token is not valid.
        """
        try:
            header, payload, _ = JWS.decode(token, key)
            decoded_token = cls(header=header, payload=payload)
            decoded_token.validate()
        except exceptions.JWTDecodeException:
            logger.debug("Error decoding token")
            raise
        except exceptions.JWTValidateException as e:
            logger.debug("Error validating token: %s", e)
            raise
        else:
            logger.debug("Decoded token: %s", decoded_token)

        return decoded_token

    def validate(self, validators: t.Optional[list[claims.ClaimValidator]] = None, **claims: t.Any) -> None:
        """Validate the token claims.

        It validates all the default claims in the payload in the following order:
        - Issuer (iss)
        - Subject (sub)
        - Audience (aud)
        - Expiration time (exp)
        - Not before (nbf)
        - Issued at (iat)
        - JWT ID (jti)

        Once all the default claims are validated, it validates the runs custom validators.

        If any of the claims is not valid, an exception is raised.

        :param validators: Custom validators to run.
        :param claims: Claims values used to validate.
        :raises JWTValidateException: If any of the claims is not valid.
        """
        invalid_claims = []

        for validator in [*VALIDATORS, *(validators or [])]:
            try:
                validator(self.payload, claims).validate()
            except exceptions.JWTClaimValidateException as e:
                logger.debug("Claim '%s' is not valid", e.claim)
                invalid_claims.append(e.claim)

        if invalid_claims:
            raise exceptions.JWTValidateException(f"Invalid claims ({', '.join(invalid_claims)})")

    def to_dict(self) -> dict[str, t.Any]:
        """Return the JWT as a dictionary.

        :return: JWT as a dictionary.
        """
        return {"header": self.header.to_dict(), "payload": self.payload.to_dict()}



import functools
from unittest.mock import AsyncMock, MagicMock, call, patch

import pytest

from flama import Flama, endpoints, exceptions, types, websockets
from flama.endpoints import WebSocketEndpoint
from flama.routing.routes.base import BaseRoute
from flama.routing.routes.websocket import WebSocketEndpointWrapper, WebSocketFunctionWrapper, WebSocketRoute


class TestCaseHTTPFunctionWrapper:
    @pytest.fixture(scope="function")
    def app(self):
        app = MagicMock(spec=Flama)
        app.router = MagicMock(resolve_route=MagicMock(side_effect=lambda x: (AsyncMock(), x)))
        app.injector = MagicMock(
            inject=AsyncMock(side_effect=lambda x, y: functools.partial(x, y["websocket"], y["websocket_message"]))
        )
        return app

    @pytest.fixture(scope="function")
    def endpoint(self):
        async def foo(websocket: websockets.WebSocket, data: types.Data):
            await websocket.send_json({"foo": "bar"})

        return WebSocketFunctionWrapper(foo)

    @pytest.fixture(scope="function")
    def websocket(self):
        return MagicMock(spec=websockets.WebSocket)

    async def test_call(self, app, endpoint, websocket):
        scope, receive, send = (
            types.Scope({"app": app, "path": "/", "root_app": app, "type": "websocket", "websocket": websocket}),
            AsyncMock(),
            AsyncMock(),
        )

        with patch("flama.websockets.WebSocket", return_value=websocket):
            await endpoint(scope, receive, send)

            assert websocket.send_json.call_args_list == [call({"foo": "bar"})]


class TestCaseHTTPEndpointWrapper:
    @pytest.fixture(scope="function")
    def app(self):
        app = MagicMock(spec=Flama)
        app.router = MagicMock(resolve_route=MagicMock(side_effect=lambda x: (AsyncMock(), x)))
        app.injector = MagicMock(
            inject=AsyncMock(side_effect=lambda x, y: functools.partial(x, y["websocket"], y["websocket_message"]))
        )
        return app

    @pytest.fixture(scope="function")
    def endpoint(self):
        class Endpoint(endpoints.WebSocketEndpoint):
            async def on_receive(self, websocket: websockets.WebSocket, data: types.Data) -> None:
                await websocket.send_json({"foo": "bar"})

        return WebSocketEndpointWrapper(Endpoint)

    @pytest.fixture(scope="function")
    def websocket(self):
        return MagicMock(spec=websockets.WebSocket)

    async def test_call(self, app, endpoint, websocket):
        scope, receive, send = (
            types.Scope({"app": app, "path": "/", "root_app": app, "type": "websocket", "websocket": websocket}),
            AsyncMock(),
            AsyncMock(),
        )

        websocket.is_connected = False
        websocket.receive = AsyncMock(return_value={})
        with patch("flama.websockets.WebSocket", return_value=websocket):
            await endpoint(scope, receive, send)


class TestCaseWebsocketRoute:
    @pytest.fixture(scope="function")
    def route(self):
        return WebSocketRoute("/", lambda: None, name="foo", include_in_schema=False)

    @pytest.fixture(scope="function")
    def endpoint(self, request):
        if request.param == "function":

            def foo(): ...

            return foo

        elif request.param == "endpoint":

            class FooEndpoint(WebSocketEndpoint):
                async def on_receive(self, websocket: websockets.WebSocket, data: types.Data) -> None:
                    await websocket.send_text("foo")

            return FooEndpoint

    @pytest.mark.parametrize(
        ["endpoint", "name", "wrapper", "exception"],
        (
            pytest.param("function", "foo", WebSocketFunctionWrapper, None, id="function"),
            pytest.param("endpoint", "FooEndpoint", WebSocketEndpointWrapper, None, id="endpoint"),
            pytest.param(
                None,
                None,
                None,
                exceptions.ApplicationError("Endpoint must be a callable or a WebSocketEndpoint subclass"),
                id="error_handler",
            ),
        ),
        indirect=["endpoint", "exception"],
    )
    def test_init(self, endpoint, name, wrapper, exception):
        with exception:
            route = WebSocketRoute("/", endpoint, include_in_schema=False)

            assert route.path == "/"
            assert isinstance(route.app, wrapper)
            assert route.endpoint == endpoint
            assert route.name == name
            assert route.include_in_schema is False

    @pytest.mark.parametrize(
        ["scope_type", "handle_call"],
        (
            pytest.param("websocket", True, id="websocket"),
            pytest.param("wrong", False, id="wrong"),
        ),
    )
    async def test_call(self, route, asgi_scope, asgi_receive, asgi_send, scope_type, handle_call):
        scope = types.Scope({**asgi_scope, "type": scope_type})
        route_scope = types.Scope({"foo": "bar"})
        handle = AsyncMock()
        expected_calls = [call(types.Scope({**scope, **route_scope}), asgi_receive, asgi_send)] if handle_call else []

        with patch.object(route, "handle", new=handle), patch.object(route, "route_scope", return_value=route_scope):
            await route(scope, asgi_receive, asgi_send)

        assert handle.call_args_list == expected_calls

    def test_eq(self):
        def foo(): ...

        assert WebSocketRoute("/", foo, name="foo") == WebSocketRoute("/", foo, name="foo")
        assert WebSocketRoute("/", foo, name="foo") != WebSocketRoute("/", foo, name="bar")

    @pytest.mark.parametrize(
        ["endpoint", "result"],
        (
            pytest.param("function", False, id="function"),
            pytest.param("endpoint", True, id="endpoint"),
        ),
        indirect=["endpoint"],
    )
    def test_is_endpoint(self, endpoint, result):
        assert WebSocketRoute.is_endpoint(endpoint) is result

    @pytest.mark.parametrize(
        ["endpoint", "expected_handlers"],
        (
            pytest.param("function", {"WEBSOCKET"}, id="function"),
            pytest.param("endpoint", {"WEBSOCKET_CONNECT", "WEBSOCKET_RECEIVE", "WEBSOCKET_DISCONNECT"}, id="endpoint"),
        ),
        indirect=["endpoint"],
    )
    def test_endpoint_handlers(self, endpoint, expected_handlers):
        result = {x: getattr(endpoint, x.lower().replace("websocket", "on"), endpoint) for x in expected_handlers}

        route = WebSocketRoute("/", endpoint)

        assert route.endpoint_handlers() == result

    @pytest.mark.parametrize(
        ["scope_type", "path_match_return", "result"],
        (
            pytest.param("websocket", BaseRoute.Match.full, BaseRoute.Match.full, id="match"),
            pytest.param("websocket", BaseRoute.Match.none, BaseRoute.Match.none, id="no_match"),
            pytest.param("http", None, BaseRoute.Match.none, id="wrong_scope_type"),
        ),
    )
    def test_match(self, scope_type, path_match_return, result, asgi_scope):
        def foo(): ...

        route = WebSocketRoute("/", foo)

        asgi_scope["type"] = scope_type

        with patch.object(BaseRoute, "match", return_value=path_match_return):
            assert route.match(asgi_scope) == result



import asyncio
from unittest.mock import AsyncMock, MagicMock, call, patch

import pytest

from flama import Flama, types
from flama.client import Client, LifespanContextManager
from flama.models.modules import ModelsModule


@pytest.fixture(scope="function")
def app():
    app_mock = AsyncMock(spec=Flama)
    app_mock.models = MagicMock(spec=ModelsModule)
    return app_mock


class TestCaseLifespanContextManager:
    @pytest.fixture(scope="function")
    def lifespan_context_manager(self, app) -> LifespanContextManager:
        return LifespanContextManager(app=app)

    @pytest.mark.parametrize(
        ["exception"],
        (
            pytest.param(None, id="ok"),
            pytest.param(Exception("foo"), id="exception"),
        ),
        indirect=["exception"],
    )
    async def test_startup(self, exception, lifespan_context_manager):
        lifespan_context_manager._exception = exception.exception
        lifespan_context_manager._startup_complete = MagicMock(spec=asyncio.Event)
        with exception:
            await lifespan_context_manager._startup()
            assert lifespan_context_manager._startup_complete.wait.await_args_list == [call()]

    async def test_shutdown(self, lifespan_context_manager):
        lifespan_context_manager._shutdown_complete = MagicMock(spec=asyncio.Event)
        await lifespan_context_manager._shutdown()
        assert lifespan_context_manager._shutdown_complete.wait.await_args_list == [call()]

    async def test_receive(self, lifespan_context_manager):
        lifespan_context_manager._receive_queue.get = AsyncMock()

        await lifespan_context_manager._receive()

        assert lifespan_context_manager._receive_queue.get.await_args_list == [call()]

    @pytest.mark.parametrize(
        ["message", "startup", "shutdown"],
        (
            pytest.param(types.Message({"type": "lifespan.startup.complete"}), True, False, id="startup"),
            pytest.param(types.Message({"type": "lifespan.shutdown.complete"}), False, True, id="shutdown"),
            pytest.param(types.Message({"type": "foo"}), False, False, id="otherwise"),
        ),
    )
    async def test_send(self, message, startup, shutdown, lifespan_context_manager):
        expected_startup = [call()] if startup else []
        expected_shutdown = [call()] if shutdown else []

        lifespan_context_manager._startup_complete.set = MagicMock()
        lifespan_context_manager._shutdown_complete.set = MagicMock()

        await lifespan_context_manager._send(message)

        assert lifespan_context_manager._startup_complete.set.call_args_list == expected_startup
        assert lifespan_context_manager._shutdown_complete.set.call_args_list == expected_shutdown

    @pytest.mark.parametrize(
        ["exception"],
        (
            pytest.param(None, id="ok"),
            pytest.param(Exception("foo"), id="exception"),
        ),
        indirect=["exception"],
    )
    async def test_app_task(self, exception, lifespan_context_manager):
        lifespan_context_manager._startup_complete.set = MagicMock()
        lifespan_context_manager._shutdown_complete.set = MagicMock()

        if exception:
            lifespan_context_manager.app.side_effect = exception.exception

        with exception:
            await lifespan_context_manager._app_task()
            assert lifespan_context_manager.app.await_args_list == [
                call(
                    types.Scope({"type": "lifespan"}), lifespan_context_manager._receive, lifespan_context_manager._send
                )
            ]

        if exception:
            assert lifespan_context_manager._exception == exception.exception
            assert lifespan_context_manager._startup_complete.set.call_args_list == [call()]
            assert lifespan_context_manager._shutdown_complete.set.call_args_list == [call()]

    @pytest.mark.parametrize(
        ["startup_mock", "shutdown_mock", "exception"],
        (
            pytest.param(
                AsyncMock(),
                AsyncMock(),
                None,
                id="ok",
            ),
            pytest.param(
                AsyncMock(side_effect=Exception("foo")),
                AsyncMock(),
                Exception("foo"),
                id="startup_exception",
            ),
            pytest.param(
                AsyncMock(),
                AsyncMock(side_effect=Exception("foo")),
                Exception("foo"),
                id="shutdown_exception",
            ),
        ),
        indirect=["exception"],
    )
    async def test_context(self, lifespan_context_manager, startup_mock, shutdown_mock, exception):
        with (
            exception,
            patch.object(lifespan_context_manager, "_app_task"),
            patch.object(lifespan_context_manager, "_startup", startup_mock),
            patch.object(lifespan_context_manager, "_shutdown", shutdown_mock),
        ):
            async with lifespan_context_manager:
                assert lifespan_context_manager._app_task.call_args_list == [call()]
                assert lifespan_context_manager._startup.await_args_list == [call()]

                if exception:
                    assert lifespan_context_manager._stop_app.await_args_list == [call()]

            assert lifespan_context_manager._app_task.call_args_list == [call(), call()]
            assert lifespan_context_manager._shutdown.await_args_list == [call()]


class TestCaseClient:
    def test_init_models(self, app):
        with patch("flama.client.Flama", return_value=app), patch("importlib.metadata.version", return_value="x.y.z"):
            client = Client(models=[("foo", "/foo/", "model_foo.flm"), ("bar", "/bar/", "model_bar.flm")])

        assert client.app == app
        assert client.lifespan
        assert client.lifespan.app == app
        assert client.models == {"foo": "/foo/", "bar": "/bar/"}
        assert app.models.add_model.call_args_list == [
            call("/foo/", "model_foo.flm", "foo"),
            call("/bar/", "model_bar.flm", "bar"),
        ]

    def test_init_no_app(self):
        with patch("importlib.metadata.version", return_value="x.y.z"):
            client = Client()

        assert client.lifespan is None

    async def test_context_app(self, app):
        client = Client(app=app)
        client.lifespan = MagicMock(spec=LifespanContextManager)

        with patch("httpx.AsyncClient.__aenter__") as aenter_mock, patch("httpx.AsyncClient.__aexit__") as aexit_mock:
            async with client:
                assert aenter_mock.await_args_list == [call()]
                assert client.lifespan.__aenter__.await_args_list == [call()]

            assert client.lifespan.__aexit__.await_args_list == [call(None, None, None)]
            assert aexit_mock.await_args_list == [call(None, None, None)]

    async def test_context_no_app(self):
        client = Client()

        with patch("httpx.AsyncClient.__aenter__") as aenter_mock, patch("httpx.AsyncClient.__aexit__") as aexit_mock:
            async with client:
                assert aenter_mock.await_args_list == [call()]

            assert aexit_mock.await_args_list == [call(None, None, None)]

    async def test_model_request(self, app):
        with (
            patch("flama.client.Flama", return_value=app),
            patch("builtins.super"),
            patch("importlib.metadata.version", return_value="x.y.z"),
        ):
            client = Client(models=[("foo", "/foo/", "model_foo.flm")])

        with patch.object(client, "request") as request_mock:
            await client.model_request("foo", "GET", "/")

            assert request_mock.call_args_list == [call("GET", "/foo/")]



import codecs
import importlib.metadata
import io
import typing as t

from flama import exceptions
from flama.serialize import types
from flama.serialize.base import Serializer

try:
    import torch  # type: ignore
except Exception:  # pragma: no cover
    torch = None  # type: ignore[misc, assignment]

if t.TYPE_CHECKING:
    from flama.types import JSONSchema


class PyTorchSerializer(Serializer):
    lib = types.Framework.torch

    def dump(self, obj: t.Any, **kwargs) -> bytes:
        if torch is None:  # noqa
            raise exceptions.FrameworkNotInstalled("pytorch")

        buffer = io.BytesIO()
        torch.jit.save(torch.jit.script(obj), buffer, **kwargs)
        buffer.seek(0)
        return codecs.encode(buffer.read(), "base64")

    def load(self, model: bytes, **kwargs) -> t.Any:
        if torch is None:  # noqa
            raise exceptions.FrameworkNotInstalled("pytorch")

        return torch.jit.load(io.BytesIO(codecs.decode(model, "base64")), **kwargs)

    def info(self, model: t.Any) -> t.Optional["JSONSchema"]:
        return {
            "modules": [str(x) for x in model.modules()],
            "parameters": {k: str(v) for k, v in model.named_parameters()},
            "state": {k: v.tolist() if hasattr(v, "tolist") else v for k, v in model.state_dict().items()},
        }

    def version(self) -> str:
        try:
            return importlib.metadata.version("torch")
        except Exception:  # noqa
            raise exceptions.FrameworkNotInstalled("pytorch")



import decimal
import re
import uuid
from decimal import Decimal

import pytest

from flama.url import (
    URL,
    DecimalSerializer,
    FloatSerializer,
    IntegerSerializer,
    Path,
    StringSerializer,
    UUIDSerializer,
    _BuildResult,
    _Fragment,
    _Match,
    _MatchResult,
)


class TestCaseFragment:
    @pytest.mark.parametrize(
        ["fragment", "serializer", "type", "name", "exception"],
        (
            pytest.param("", None, "constant", None, None, id="empty"),
            pytest.param("{foo}", StringSerializer(), "str", "foo", None, id="no_type"),
            pytest.param("{foo:str}", StringSerializer(), "str", "foo", None, id="str"),
            pytest.param("{foo:int}", IntegerSerializer(), "int", "foo", None, id="int"),
            pytest.param("{foo:float}", FloatSerializer(), "float", "foo", None, id="float"),
            pytest.param("{foo:decimal}", DecimalSerializer(), "decimal", "foo", None, id="decimal"),
            pytest.param("{foo:uuid}", UUIDSerializer(), "uuid", "foo", None, id="uuid"),
            pytest.param("{foo:bar}", None, None, None, ValueError("Unknown path serializer 'bar'"), id="unknown_type"),
        ),
        indirect=["exception"],
    )
    def test_init(self, fragment, serializer, type, name, exception):
        with exception:
            result = _Fragment.build(fragment)

            assert result.value == fragment
            assert result.type == type
            assert getattr(result, "serializer", None) == serializer
            assert getattr(result, "name", None) == name


class TestCasePath:
    @pytest.mark.parametrize(
        ["path", "template", "regex", "parameters", "exception"],
        (
            pytest.param(
                "foo/bar",
                None,
                None,
                None,
                ValueError("Path must starts with '/'"),
                id="wrong",
            ),
            pytest.param(
                "",
                "",
                re.compile(r"^(?P<__matched__>)(?P<__unmatched__>.*)$"),
                {},
                None,
                id="empty",
            ),
            pytest.param(
                "/",
                "/",
                re.compile(r"^(?P<__matched__>/)(?P<__unmatched__>.*)$"),
                {},
                None,
                id="no_params",
            ),
            pytest.param(
                "/{foo}/",
                "/{foo}/",
                re.compile(r"^(?P<__matched__>/(?P<foo>[^/]+)/)(?P<__unmatched__>.*)$"),
                {"foo": _Fragment.build("{foo}")},
                None,
                id="no_type",
            ),
            pytest.param(
                "/{foo:str}/",
                "/{foo}/",
                re.compile(r"^(?P<__matched__>/(?P<foo>[^/]+)/)(?P<__unmatched__>.*)$"),
                {"foo": _Fragment.build("{foo:str}")},
                None,
                id="str",
            ),
            pytest.param(
                "/{foo:int}/",
                "/{foo}/",
                re.compile(r"^(?P<__matched__>/(?P<foo>-?[0-9]+)/)(?P<__unmatched__>.*)$"),
                {"foo": _Fragment.build("{foo:int}")},
                None,
                id="int",
            ),
            pytest.param(
                "/{foo:float}/",
                "/{foo}/",
                re.compile(r"^(?P<__matched__>/(?P<foo>-?[0-9]+(.[0-9]+)?)/)(?P<__unmatched__>.*)$"),
                {"foo": _Fragment.build("{foo:float}")},
                None,
                id="float",
            ),
            pytest.param(
                "/{foo:decimal}/",
                "/{foo}/",
                re.compile(r"^(?P<__matched__>/(?P<foo>-?[0-9]+(.[0-9]+)?)/)(?P<__unmatched__>.*)$"),
                {"foo": _Fragment.build("{foo:decimal}")},
                None,
                id="decimal",
            ),
            pytest.param(
                "/{foo:uuid}/",
                "/{foo}/",
                re.compile(
                    r"^(?P<__matched__>/(?P<foo>[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12})/)(?P<__unmatched__>.*)$"
                ),
                {"foo": _Fragment.build("{foo:uuid}")},
                None,
                id="uuid",
            ),
            pytest.param(
                "/{foo:bar}",
                None,
                None,
                None,
                ValueError("Unknown path serializer 'bar'"),
                id="unknown_type",
            ),
        ),
        indirect=["exception"],
    )
    def test_init(self, path, template, regex, parameters, exception):
        with exception:
            result = Path(path)

            assert result.path == path
            assert result._template == template
            assert result._regex == regex
            assert result._parameters == parameters

    @pytest.mark.parametrize(
        ["path", "value", "result"],
        (
            pytest.param(
                Path("/"), "/unused/path", _MatchResult(_Match.partial, {}, "/", "unused/path"), id="unmatched"
            ),
            pytest.param(Path("/"), "/", _MatchResult(_Match.exact, {}, "/", None), id="no_params"),
            pytest.param(
                Path("/{foo}/"), "/bar/", _MatchResult(_Match.exact, {"foo": "bar"}, "/bar/", None), id="no_type"
            ),
            pytest.param(
                Path("/{foo:str}/"), "/bar/", _MatchResult(_Match.exact, {"foo": "bar"}, "/bar/", None), id="str"
            ),
            pytest.param(
                Path("/{foo:int}/"), "/1/", _MatchResult(_Match.exact, {"foo": 1}, "/1/", None), id="int_positive"
            ),
            pytest.param(
                Path("/{foo:int}/"), "/-1/", _MatchResult(_Match.exact, {"foo": -1}, "/-1/", None), id="int_negative"
            ),
            pytest.param(Path("/{foo:int}/"), "/foo/", _MatchResult(_Match.none, None, None, None), id="int_fail"),
            pytest.param(
                Path("/{foo:float}/"),
                "/1.0/",
                _MatchResult(_Match.exact, {"foo": 1.0}, "/1.0/", None),
                id="float_positive",
            ),
            pytest.param(
                Path("/{foo:float}/"),
                "/1/",
                _MatchResult(_Match.exact, {"foo": 1.0}, "/1/", None),
                id="float_positive_no_decimals",
            ),
            pytest.param(
                Path("/{foo:float}/"),
                "/-1.0/",
                _MatchResult(_Match.exact, {"foo": -1.0}, "/-1.0/", None),
                id="float_negative",
            ),
            pytest.param(
                Path("/{foo:float}/"),
                "/-1/",
                _MatchResult(_Match.exact, {"foo": -1.0}, "/-1/", None),
                id="float_negative_no_decimals",
            ),
            pytest.param(Path("/{foo:float}/"), "/foo/", _MatchResult(_Match.none, None, None, None), id="float_fail"),
            pytest.param(
                Path("/{foo:decimal}/"),
                "/1.0/",
                _MatchResult(_Match.exact, {"foo": Decimal("1.0")}, "/1.0/", None),
                id="decimal_positive",
            ),
            pytest.param(
                Path("/{foo:decimal}/"),
                "/1/",
                _MatchResult(_Match.exact, {"foo": Decimal("1.0")}, "/1/", None),
                id="decimal_positive_no_decimals",
            ),
            pytest.param(
                Path("/{foo:decimal}/"),
                "/-1.0/",
                _MatchResult(_Match.exact, {"foo": Decimal("-1.0")}, "/-1.0/", None),
                id="decimal_negative",
            ),
            pytest.param(
                Path("/{foo:decimal}/"),
                "/-1/",
                _MatchResult(_Match.exact, {"foo": Decimal("-1.0")}, "/-1/", None),
                id="decimal_negative_no_decimals",
            ),
            pytest.param(
                Path("/{foo:decimal}/"), "/foo/", _MatchResult(_Match.none, None, None, None), id="decimal_fail"
            ),
            pytest.param(
                Path("/{foo:uuid}/"),
                "/83a8e611-525c-4d30-9bbf-f2c142606a3d/",
                _MatchResult(
                    _Match.exact,
                    {"foo": uuid.UUID("83a8e611-525c-4d30-9bbf-f2c142606a3d")},
                    "/83a8e611-525c-4d30-9bbf-f2c142606a3d/",
                    None,
                ),
                id="uuid",
            ),
            pytest.param(
                Path("/{foo:uuid}/"),
                "/f2c142606a3d/",
                _MatchResult(_Match.none, None, None, None),
                id="uuid_fail",
            ),
        ),
    )
    def test_match(self, path, value, result):
        assert path.match(value) == result

    @pytest.mark.parametrize(
        ["path", "result"],
        (
            pytest.param(Path("/"), {}, id="no_params"),
            pytest.param(Path("/{foo}/"), {"foo": str}, id="no_type"),
            pytest.param(Path("/{foo:str}/"), {"foo": str}, id="str"),
            pytest.param(Path("/{foo:int}/"), {"foo": int}, id="int"),
            pytest.param(Path("/{foo:float}/"), {"foo": float}, id="float"),
            pytest.param(Path("/{foo:decimal}/"), {"foo": decimal.Decimal}, id="decimal"),
            pytest.param(Path("/{foo:uuid}/"), {"foo": uuid.UUID}, id="uuid"),
        ),
    )
    def test_parameters(self, path, result):
        assert path.parameters == result

    @pytest.mark.parametrize(
        ["path", "params", "result", "exception"],
        (
            pytest.param(Path("/"), {}, _BuildResult("/", {}), None, id="no_params"),
            pytest.param(Path("/{foo}/"), {}, None, ValueError("Wrong params, expected: 'foo'."), id="wrong_params"),
            pytest.param(
                Path("/{foo:int}/"), {"foo": 1, "bar": 1}, _BuildResult("/1/", {"bar": 1}), None, id="remaining_params"
            ),
            pytest.param(Path("/{foo}/"), {"foo": "bar"}, _BuildResult("/bar/", {}), None, id="no_type"),
            pytest.param(Path("/{foo:str}/"), {"foo": "bar"}, _BuildResult("/bar/", {}), None, id="str"),
            pytest.param(Path("/{foo:int}/"), {"foo": 1}, _BuildResult("/1/", {}), None, id="int_positive"),
            pytest.param(Path("/{foo:int}/"), {"foo": -1}, _BuildResult("/-1/", {}), None, id="int_negative"),
            pytest.param(Path("/{foo:float}/"), {"foo": 1.1}, _BuildResult("/1.1/", {}), None, id="float_positive"),
            pytest.param(
                Path("/{foo:float}/"), {"foo": 1.0}, _BuildResult("/1/", {}), None, id="float_positive_no_decimals"
            ),
            pytest.param(Path("/{foo:float}/"), {"foo": -1.1}, _BuildResult("/-1.1/", {}), None, id="float_negative"),
            pytest.param(
                Path("/{foo:float}/"), {"foo": -1.0}, _BuildResult("/-1/", {}), None, id="float_negative_no_decimals"
            ),
            pytest.param(
                Path("/{foo:decimal}/"), {"foo": Decimal("1.1")}, _BuildResult("/1.1/", {}), None, id="decimal_positive"
            ),
            pytest.param(
                Path("/{foo:decimal}/"),
                {"foo": Decimal("1.0")},
                _BuildResult("/1.0/", {}),
                None,
                id="decimal_positive_no_decimals",
            ),
            pytest.param(
                Path("/{foo:decimal}/"),
                {"foo": Decimal("-1.1")},
                _BuildResult("/-1.1/", {}),
                None,
                id="decimal_negative",
            ),
            pytest.param(
                Path("/{foo:decimal}/"),
                {"foo": Decimal("-1.0")},
                _BuildResult("/-1.0/", {}),
                None,
                id="decimal_negative_no_decimals",
            ),
            pytest.param(
                Path("/{foo:uuid}/"),
                {"foo": uuid.UUID("83a8e611-525c-4d30-9bbf-f2c142606a3d")},
                _BuildResult("/83a8e611-525c-4d30-9bbf-f2c142606a3d/", {}),
                None,
                id="uuid",
            ),
            pytest.param(
                Path("/{foo:uuid}/bar/{bar:uuid}/"),
                {
                    "foo": uuid.UUID("83a8e611-525c-4d30-9bbf-f2c142606a3d"),
                    "bar": uuid.UUID("ac187df6-d785-4aed-bcb7-84ed4a461e0c"),
                },
                _BuildResult("/83a8e611-525c-4d30-9bbf-f2c142606a3d/bar/ac187df6-d785-4aed-bcb7-84ed4a461e0c/", {}),
                None,
                id="multiple_uuid",
            ),
        ),
        indirect=["exception"],
    )
    def test_build(self, path, params, result, exception):
        with exception:
            assert path.build(**params) == result

    @pytest.mark.parametrize(
        ["path", "result"],
        (
            pytest.param("/", True, id="true"),
            pytest.param("", False, id="false"),
        ),
    )
    def test_bool(self, path, result):
        assert bool(Path(path)) is result

    @pytest.mark.parametrize(
        ["a", "b", "c", "exception"],
        (
            pytest.param(Path("/foo"), "/bar", Path("/foo/bar"), None, id="str"),
            pytest.param(Path("/foo/"), "/bar/", Path("/foo/bar/"), None, id="str_trailing_slash"),
            pytest.param(Path("/foo"), Path("/bar"), Path("/foo/bar"), None, id="path"),
            pytest.param(Path("/foo/"), Path("/bar/"), Path("/foo/bar/"), None, id="path_trailing_slash"),
            pytest.param(Path("/foo"), 1, None, TypeError("Can only concatenate str or Path to Path"), id="error"),
        ),
        indirect=["exception"],
    )
    def test_truediv(self, a, b, c, exception):
        with exception:
            assert a / b == c

    @pytest.mark.parametrize(
        ["a", "b", "c", "exception"],
        (
            pytest.param(Path("/foo"), "/bar", Path("/bar/foo"), None, id="str"),
            pytest.param(Path("/foo/"), "/bar/", Path("/bar/foo/"), None, id="str_trailing_slash"),
            pytest.param(Path("/foo"), 1, None, TypeError("Can only concatenate str or Path to Path"), id="error"),
        ),
        indirect=["exception"],
    )
    def test_rtruediv(self, a, b, c, exception):
        with exception:
            assert b / a == c

    @pytest.mark.parametrize(
        ["a", "b", "c", "exception"],
        (
            pytest.param(Path("/foo"), "/bar", Path("/foo/bar"), None, id="str"),
            pytest.param(Path("/foo/"), "/bar/", Path("/foo/bar/"), None, id="str_trailing_slash"),
            pytest.param(Path("/foo"), Path("/bar"), Path("/foo/bar"), None, id="path"),
            pytest.param(Path("/foo/"), Path("/bar/"), Path("/foo/bar/"), None, id="path_trailing_slash"),
            pytest.param(Path("/foo"), 1, None, TypeError("Can only concatenate str or Path to Path"), id="error"),
        ),
        indirect=["exception"],
    )
    def test_itruediv(self, a, b, c, exception):
        with exception:
            a /= b
            assert a == c


class TestCaseURL:
    @pytest.mark.parametrize(
        ["url", "components"],
        (
            pytest.param(
                "",
                {"scheme": "", "netloc": "", "path": "", "params": "", "query": "", "fragment": ""},
                id="str_empty",
            ),
            pytest.param(
                "https://www.foo.bar/foobar?foo=bar",
                {
                    "scheme": "https",
                    "netloc": "www.foo.bar",
                    "path": "/foobar",
                    "params": "",
                    "query": "foo=bar",
                    "fragment": "",
                },
                id="str_basic",
            ),
            pytest.param(
                "https://user:pass@www.foo.bar:8000/foobar?foo=bar#barfoo",
                {
                    "scheme": "https",
                    "netloc": "user:pass@www.foo.bar:8000",
                    "path": "/foobar",
                    "params": "",
                    "query": "foo=bar",
                    "fragment": "barfoo",
                },
                id="str_full",
            ),
            pytest.param(
                URL(""),
                {"scheme": "", "netloc": "", "path": "", "params": "", "query": "", "fragment": ""},
                id="url_empty",
            ),
            pytest.param(
                URL("https://www.foo.bar/foobar?foo=bar"),
                {
                    "scheme": "https",
                    "netloc": "www.foo.bar",
                    "path": "/foobar",
                    "params": "",
                    "query": "foo=bar",
                    "fragment": "",
                },
                id="url_basic",
            ),
            pytest.param(
                URL("https://user:pass@www.foo.bar:8000/foobar?foo=bar#barfoo"),
                {
                    "scheme": "https",
                    "netloc": "user:pass@www.foo.bar:8000",
                    "path": "/foobar",
                    "params": "",
                    "query": "foo=bar",
                    "fragment": "barfoo",
                },
                id="url_full",
            ),
        ),
    )
    def test_url(self, url, components):
        assert URL(url) == URL(**components)



import pytest

from flama import endpoints, http, injection, websockets
from flama.applications import Flama
from flama.client import Client


class Puppy:
    name = "Canna"


class Owner:
    name = "Perdy"

    def __init__(self, puppy: Puppy):
        self.puppy = puppy


class Unknown:
    pass


class Foo:
    name = "Foo"


class TestCaseComponentsInjection:
    @pytest.fixture(scope="class")
    def puppy_component(self):
        class PuppyComponent(injection.Component):
            def resolve(self) -> Puppy:
                return Puppy()

        return PuppyComponent()

    @pytest.fixture(scope="class")
    def unknown_param_component(self):
        class UnknownParamComponent(injection.Component):
            def resolve(self, foo: Unknown) -> Foo:
                return Foo()

        return UnknownParamComponent()

    @pytest.fixture(scope="class")
    def owner_component(self):
        class OwnerComponent(injection.Component):
            def resolve(self, puppy: Puppy) -> Owner:
                return Owner(puppy=puppy)

        return OwnerComponent()

    @pytest.fixture(scope="function", autouse=True)
    def add_components(self, app, puppy_component, owner_component, unknown_param_component):
        for component in (puppy_component, owner_component, unknown_param_component):
            app.add_component(component)

    @pytest.fixture(scope="function", autouse=True)
    def add_endpoints(self, app):
        @app.route("/http-view/")
        async def puppy_http_view(puppy: Puppy):
            return http.JSONResponse({"puppy": puppy.name})

        @app.route("/http-endpoint/", methods=["GET"])
        class PuppyHTTPEndpoint(endpoints.HTTPEndpoint):
            async def get(self, puppy: Puppy):
                return http.JSONResponse({"puppy": puppy.name})

        @app.websocket_route("/websocket-view/")
        async def puppy_websocket_view(session: websockets.WebSocket, puppy: Puppy):
            await session.accept()
            await session.send_json({"puppy": puppy.name})
            await session.close()

        @app.route("/nested-component/")
        async def nested_component_view(owner: Owner):
            return {"name": owner.name, "puppy": {"name": owner.puppy.name}}

        @app.route("/unknown-component/")
        def unknown_component_view(unknown: Unknown):
            return http.JSONResponse({"foo": "bar"})

        @app.route("/unknown-param-in-component/")
        def unknown_param_in_component_view(foo: Foo):
            return http.JSONResponse({"foo": "bar"})

    @pytest.mark.parametrize(
        ["url", "method", "params", "status_code", "result", "exception"],
        (
            pytest.param("/http-view/", "get", None, 200, {"puppy": "Canna"}, None, id="http_view"),
            pytest.param("/http-endpoint/", "get", None, 200, {"puppy": "Canna"}, None, id="http_endpoint"),
            pytest.param(
                "/http-endpoint/",
                "websocket",
                None,
                None,
                {"puppy": "Canna"},
                None,
                marks=pytest.mark.skip(
                    reason="Cannot test websockets with current client"
                ),  # CAVEAT: Client doesn't support websockets
                id="websocket",
            ),
            pytest.param(
                "/nested-component/",
                "get",
                None,
                200,
                {"name": "Perdy", "puppy": {"name": "Canna"}},
                None,
                id="nested_component",
            ),
            pytest.param(
                "/unknown-component/",
                "get",
                None,
                None,
                None,
                (
                    injection.ComponentNotFound,
                    "No component able to handle parameter 'unknown' for function 'unknown_component_view'",
                ),
                id="unknown_component",
            ),
            pytest.param(
                "/unknown-param-in-component/",
                "get",
                None,
                None,
                None,
                (
                    injection.ComponentNotFound,
                    "No component able to handle parameter 'foo' in component 'UnknownParamComponent' for function "
                    "'unknown_param_in_component_view'",
                ),
                id="unknown_param_in_component",
            ),
        ),
        indirect=["exception"],
    )
    async def test_injection(self, client, url, method, params, status_code, result, exception):
        with exception:
            if method == "websocket":
                with client.websocket_connect(url) as websocket:
                    assert websocket.receive_json() == result
            else:
                response = await client.request(method, url, params=params)

                assert response.status_code == status_code
                assert response.json() == result

    async def test_unhandled_component(self):
        class UnhandledComponent(injection.Component):
            def resolve(self):
                pass

        app = Flama(components=[UnhandledComponent()])

        @app.route("/")
        def foo(unknown: Unknown):
            return http.JSONResponse({"foo": "bar"})

        with pytest.raises(
            injection.ComponentError,
            match="Component 'UnhandledComponent' must include a return annotation on the 'resolve' method, "
            "or override 'can_handle_parameter'",
        ):
            async with Client(app) as client:
                await client.request("get", "/")



from flama.applications import *  # noqa
from flama.background import *  # noqa
from flama.cli import *  # noqa
from flama.injection.components import *  # noqa
from flama.modules import *  # noqa
from flama.serialize import *  # noqa



import enum
import typing as t

from flama import compat

if t.TYPE_CHECKING:
    from flama import Flama


__all__ = ["AppStatus", "is_flama_instance"]


class AppStatus(enum.Enum):
    NOT_STARTED = enum.auto()
    STARTING = enum.auto()
    READY = enum.auto()
    SHUTTING_DOWN = enum.auto()
    SHUT_DOWN = enum.auto()
    FAILED = enum.auto()


def is_flama_instance(
    obj: t.Any,
) -> compat.TypeGuard["Flama"]:  # PORT: Replace compat when stop supporting 3.9
    """Checks if an object is an instance of Flama.

    :param obj: The object to check.
    :return: True if the object is an instance of Flama, False otherwise.
    """
    from flama import Flama

    return isinstance(obj, Flama)



from unittest.mock import AsyncMock, call, patch

import pytest
import starlette.websockets

from flama import websockets


class TestCaseWebSocket:
    @pytest.fixture
    def websocket(self, asgi_scope, asgi_receive, asgi_send):
        asgi_scope["type"] = "websocket"
        return websockets.WebSocket(asgi_scope, asgi_receive, asgi_send)

    @pytest.mark.skip(reason="Cannot test websockets with current client")  # CAVEAT: Client doesn't support websockets
    @pytest.mark.parametrize(
        ["encoding", "send_method", "data", "expected_result", "exception"],
        (
            pytest.param("bytes", "send_bytes", b"foo", {"bytes": b"foo", "type": "websocket.send"}, None, id="bytes"),
            pytest.param(
                "bytes",
                "send_text",
                "foo",
                {"code": 1003, "type": "websocket.close", "reason": ""},
                ValueError("Foo"),
                id="bytes_wrong",
            ),
            pytest.param("text", "send_text", "foo", {"text": "foo", "type": "websocket.send"}, None, id="text"),
            pytest.param(
                "text",
                "send_bytes",
                b"foo",
                {"code": 1003, "type": "websocket.close", "reason": ""},
                ValueError("Foo"),
                id="text_wrong",
            ),
            pytest.param(
                "json",
                "send_json",
                {"foo": "bar"},
                {"text": '{"foo":"bar"}', "type": "websocket.send"},
                None,
                id="json",
            ),
            pytest.param(
                "json",
                "send_bytes",
                b"foo",
                {"code": 1003, "type": "websocket.close", "reason": ""},
                ValueError("Foo"),
                id="json_wrong",
            ),
        ),
        indirect=["exception"],
    )
    def test_receive(self, app, client, encoding, send_method, data, expected_result, exception):
        @app.websocket_route("/")
        async def foo_websocket(websocket: websockets.WebSocket):
            try:
                await websocket.accept()
                if exception.exception:
                    raise exception.exception
                await getattr(websocket, f"send_{encoding or 'bytes'}")(data)
            except Exception as e:
                await websocket.close(1003)
                raise e from None
            else:
                await websocket.close()

        with exception, client.websocket_connect("/") as ws:
            getattr(ws, send_method)(data)
            result = ws.receive()

        assert result == expected_result

    def test_is_connecting(self, websocket):
        websocket.client_state = starlette.websockets.WebSocketState.CONNECTING

        assert websocket.is_connecting

    def test_is_connected(self, websocket):
        websocket.client_state = starlette.websockets.WebSocketState.CONNECTED

        assert websocket.is_connected

    def test_is_disconnected(self, websocket):
        websocket.client_state = starlette.websockets.WebSocketState.DISCONNECTED

        assert websocket.is_disconnected


class TestCaseClose:
    @pytest.fixture
    def close(self):
        return websockets.Close()

    async def test_call(self, close, asgi_scope, asgi_receive, asgi_send):
        with patch.object(starlette.websockets.WebSocketClose, "__call__", new=AsyncMock()) as super_call_mock:
            await close(asgi_scope, asgi_receive, asgi_send)

            assert super_call_mock.call_args_list == [call(asgi_scope, asgi_receive, asgi_send)]



from unittest.mock import AsyncMock, call, patch

import pytest

from flama.ddd.repositories.base import BaseRepository
from flama.ddd.workers.base import AbstractWorker, BaseWorker
from flama.exceptions import ApplicationError


@pytest.fixture(scope="function")
def repository():
    class FooRepository(BaseRepository): ...

    return FooRepository


class TestCaseAbstractWorker:
    @pytest.fixture(scope="function")
    def worker(self):
        class FooWorker(AbstractWorker):
            async def begin(self): ...

            async def end(self, *, rollback: bool = False): ...

            async def commit(self): ...

            async def rollback(self): ...

        return FooWorker()

    def test_app(self, app, worker):
        with pytest.raises(ApplicationError, match="Worker not initialized"):
            worker.app

        worker.app = app

        assert worker.app == app

        del worker.app

        with pytest.raises(ApplicationError, match="Worker not initialized"):
            worker.app

    async def test_async_context(self, app, worker):
        worker.app = app

        with patch.multiple(worker, begin=AsyncMock(), end=AsyncMock()):
            async with worker:
                assert worker.begin.await_args_list == [call()]
                assert worker.end.await_args_list == []

            assert worker.begin.await_args_list == [call()]
            assert worker.end.await_args_list == [call(rollback=False)]


class TestCaseBaseWorker:
    @pytest.fixture(scope="function")
    def worker(self, repository):
        class FooWorker(BaseWorker):
            foo: repository

            async def set_up(self): ...

            async def tear_down(self, *, rollback: bool = False): ...

            async def repository_params(self):
                return [], {}

            async def begin(self): ...

            async def end(self, *, rollback: bool = False): ...

            async def commit(self): ...

            async def rollback(self): ...

        return FooWorker()

    def test_new(self, worker, repository):
        assert hasattr(worker, "_repositories")
        assert worker._repositories == {"foo": repository}



import typing as t
from collections import namedtuple

import marshmallow
import pydantic
import pytest
import typesystem
import typesystem.fields
from pytest import param

from flama import schemas
from flama.pagination import paginator
from tests.asserts import assert_recursive_contains


@pytest.fixture(scope="function")
def app(app):
    paginator.schemas = {}
    return app


@pytest.fixture(scope="function")
def output_schema(app):
    if app.schema.schema_library.lib == pydantic:
        schema = pydantic.create_model("OutputSchema", value=(t.Optional[int], ...), __module__="pydantic.main")
        name = "pydantic.main.OutputSchema"
    elif app.schema.schema_library.lib == typesystem:
        schema = typesystem.Schema(title="OutputSchema", fields={"value": typesystem.fields.Integer(allow_null=True)})
        name = "typesystem.schemas.OutputSchema"
    elif app.schema.schema_library.lib == marshmallow:
        schema = type("OutputSchema", (marshmallow.Schema,), {"value": marshmallow.fields.Integer(allow_none=True)})
        name = "abc.OutputSchema"
    else:
        raise ValueError(f"Wrong schema lib: {app.schema.schema_library.lib}")

    return namedtuple("OutputSchema", ("schema", "name"))(schema=schema, name=name)


class TestCasePageNumberPagination:
    @pytest.fixture(scope="function", autouse=True)
    def add_endpoints(self, app, output_schema):
        @app.route("/page-number/", methods=["GET"], pagination="page_number")
        def page_number(
            **kwargs,
        ) -> t.Annotated[list[schemas.SchemaType], schemas.SchemaMetadata(output_schema.schema)]:
            return [{"value": i} for i in range(25)]

    def test_registered_schemas(self, app, output_schema):
        schemas = app.schema.schema["components"]["schemas"]
        name_prefix = output_schema.name.rsplit(".", 1)[0]

        assert set(schemas.keys()) == {
            f"{name_prefix}.OutputSchema",
            f"{name_prefix}.PageNumberPaginatedOutputSchema",
            "flama.PageNumberMeta",
            "flama.APIError",
        }

    def test_invalid_view(self, output_schema):
        with pytest.raises(TypeError, match=r"Paginated views must define \*\*kwargs param"):

            @paginator._paginate_page_number
            def invalid() -> t.Annotated[schemas.SchemaType, schemas.SchemaMetadata(output_schema.schema)]: ...

    def test_invalid_response(self):
        with pytest.raises(ValueError, match=r"Wrong schema type"):

            @paginator._paginate_page_number
            def invalid(): ...

    def test_pagination_schema_parameters(self, app):
        schema = app.schema.schema["paths"]["/page-number/"]["get"]
        parameters = schema.get("parameters", [])

        assert_recursive_contains(
            {
                "name": "count",
                "in": "query",
                "required": False,
                "schema": {"anyOf": [{"type": "boolean"}, {"type": "null"}], "default": False},
            },
            parameters[0],
        )
        assert_recursive_contains(
            {
                "name": "page",
                "in": "query",
                "required": False,
                "schema": {"anyOf": [{"type": "integer"}, {"type": "null"}]},
            },
            parameters[1],
        )
        assert_recursive_contains(
            {
                "name": "page_size",
                "in": "query",
                "required": False,
                "schema": {"anyOf": [{"type": "integer"}, {"type": "null"}]},
            },
            parameters[2],
        )

    def test_pagination_schema_return(self, app, output_schema):
        prefix, name = output_schema.name.rsplit(".", 1)
        paginated_output_schema_name = f"{prefix}.PageNumberPaginated{name}"
        response_schema = app.schema.schema["paths"]["/page-number/"]["get"]["responses"]["200"]
        component_schema = app.schema.schema["components"]["schemas"][paginated_output_schema_name]

        assert "data" in component_schema["properties"]
        assert component_schema["properties"]["data"]["items"] == {"$ref": f"#/components/schemas/{output_schema.name}"}
        assert component_schema["properties"]["data"]["type"] == "array"
        assert set(component_schema["required"]) == {"meta", "data"}
        assert component_schema["type"] == "object"

        assert response_schema == {
            "description": "Description not provided.",
            "content": {
                "application/json": {
                    "schema": {"$ref": f"#/components/schemas/{paginated_output_schema_name}"},
                }
            },
        }

    async def test_async_function(self, app, client, output_schema):
        @app.route("/page-number-async/", methods=["GET"], pagination="page_number")
        async def page_number_async(
            **kwargs,
        ) -> t.Annotated[list[schemas.SchemaType], schemas.SchemaMetadata(output_schema.schema)]:
            return [{"value": i} for i in range(25)]

        response = await client.get("/page-number-async/")
        assert response.status_code == 200, response.json()
        assert response.json() == {
            "meta": {"page": 1, "page_size": 10, "count": None},
            "data": [{"value": i} for i in range(10)],
        }

    @pytest.mark.parametrize(
        "params,status_code,expected",
        (
            param(
                {},
                200,
                {"meta": {"page_size": 10, "page": 1, "count": None}, "data": [{"value": i} for i in range(10)]},
                id="default_params",
            ),
            param(
                {"page_size": 5},
                200,
                {"meta": {"page_size": 5, "page": 1, "count": None}, "data": [{"value": i} for i in range(5)]},
                id="explicit_page_size",
            ),
            param(
                {"page": 2},
                200,
                {"meta": {"page_size": 10, "page": 2, "count": None}, "data": [{"value": i} for i in range(10, 20)]},
                id="explicit_page",
            ),
            param(
                {"page": 4, "page_size": 5},
                200,
                {"meta": {"page_size": 5, "page": 4, "count": None}, "data": [{"value": i} for i in range(15, 20)]},
                id="explicit_page_and_page_size",
            ),
            param(
                {"count": True},
                200,
                {"meta": {"page_size": 10, "page": 1, "count": 25}, "data": [{"value": i} for i in range(10)]},
                id="count",
            ),
        ),
    )
    async def test_params(self, client, params, status_code, expected):
        response = await client.get("/page-number/", params=params)
        assert response.status_code == status_code, response.json()
        assert response.json() == expected


class TestCaseLimitOffsetPagination:
    @pytest.fixture(scope="function", autouse=True)
    def add_endpoints(self, app, output_schema):
        @app.route("/limit-offset/", methods=["GET"], pagination="limit_offset")
        def limit_offset(
            **kwargs,
        ) -> t.Annotated[list[schemas.SchemaType], schemas.SchemaMetadata(output_schema.schema)]:
            return [{"value": i} for i in range(25)]

    def test_registered_schemas(self, app, output_schema):
        schemas = app.schema.schema["components"]["schemas"]
        name_prefix = output_schema.name.rsplit(".", 1)[0]

        assert set(schemas.keys()) == {
            f"{name_prefix}.OutputSchema",
            f"{name_prefix}.LimitOffsetPaginatedOutputSchema",
            "flama.LimitOffsetMeta",
            "flama.APIError",
        }

    def test_invalid_view(self, output_schema):
        with pytest.raises(TypeError, match=r"Paginated views must define \*\*kwargs param"):

            @paginator._paginate_limit_offset
            def invalid() -> t.Annotated[schemas.SchemaType, schemas.SchemaMetadata(output_schema.schema)]: ...

    def test_invalid_response(self):
        with pytest.raises(ValueError, match=r"Wrong schema type"):

            @paginator._paginate_limit_offset
            def invalid(): ...

    def test_pagination_schema_parameters(self, app):
        schema = app.schema.schema["paths"]["/limit-offset/"]["get"]
        parameters = schema.get("parameters", [])

        assert_recursive_contains(
            {
                "name": "count",
                "in": "query",
                "required": False,
                "schema": {"anyOf": [{"type": "boolean"}, {"type": "null"}], "default": False},
            },
            parameters[0],
        )
        assert_recursive_contains(
            {
                "name": "limit",
                "in": "query",
                "required": False,
                "schema": {"anyOf": [{"type": "integer"}, {"type": "null"}]},
            },
            parameters[1],
        )
        assert_recursive_contains(
            {
                "name": "offset",
                "in": "query",
                "required": False,
                "schema": {"anyOf": [{"type": "integer"}, {"type": "null"}]},
            },
            parameters[2],
        )

    def test_pagination_schema_return(self, app, output_schema):
        prefix, name = output_schema.name.rsplit(".", 1)
        paginated_output_schema_name = f"{prefix}.LimitOffsetPaginated{name}"

        response_schema = app.schema.schema["paths"]["/limit-offset/"]["get"]["responses"]["200"]
        component_schema = app.schema.schema["components"]["schemas"][paginated_output_schema_name]

        assert "data" in component_schema["properties"]
        assert component_schema["properties"]["data"]["items"] == {"$ref": f"#/components/schemas/{output_schema.name}"}
        assert component_schema["properties"]["data"]["type"] == "array"
        assert set(component_schema["required"]) == {"meta", "data"}
        assert component_schema["type"] == "object"

        assert response_schema == {
            "description": "Description not provided.",
            "content": {
                "application/json": {
                    "schema": {"$ref": f"#/components/schemas/{paginated_output_schema_name}"},
                }
            },
        }

    async def test_async_function(self, app, client, output_schema):
        @app.route("/limit-offset-async/", methods=["GET"], pagination="limit_offset")
        async def limit_offset_async(
            **kwargs,
        ) -> t.Annotated[list[schemas.SchemaType], schemas.SchemaMetadata(output_schema.schema)]:
            return [{"value": i} for i in range(25)]

        response = await client.get("/limit-offset-async/")
        assert response.status_code == 200, response.json()
        assert response.json() == {
            "meta": {"limit": 10, "offset": 0, "count": None},
            "data": [{"value": i} for i in range(10)],
        }

    @pytest.mark.parametrize(
        "params,status_code,expected",
        (
            param(
                {},
                200,
                {"meta": {"limit": 10, "offset": 0, "count": None}, "data": [{"value": i} for i in range(10)]},
                id="default_params",
            ),
            param(
                {"limit": 5},
                200,
                {"meta": {"limit": 5, "offset": 0, "count": None}, "data": [{"value": i} for i in range(5)]},
                id="explicit_limit",
            ),
            param(
                {"offset": 5},
                200,
                {"meta": {"limit": 10, "offset": 5, "count": None}, "data": [{"value": i} for i in range(5, 15)]},
                id="explicit_offset",
            ),
            param(
                {"offset": 5, "limit": 20},
                200,
                {"meta": {"limit": 20, "offset": 5, "count": None}, "data": [{"value": i} for i in range(5, 25)]},
                id="explicit_offset_and_limit",
            ),
            param(
                {"count": True},
                200,
                {"meta": {"limit": 10, "offset": 0, "count": 25}, "data": [{"value": i} for i in range(10)]},
                id="count",
            ),
        ),
    )
    async def test_params(self, client, params, status_code, expected):
        response = await client.get("/limit-offset/", params=params)
        assert response.status_code == status_code, response.json()
        assert response.json() == expected



import functools
import multiprocessing
import warnings

import pytest

from flama import concurrency


def sync_event_task(event):
    event.set()


async def async_event_task(event):
    event.set()


class TestCaseIsAsync:
    def test_function(self):
        def foo_sync(): ...

        async def foo_async(): ...

        assert not concurrency.is_async(foo_sync)
        assert concurrency.is_async(foo_async)

    def test_callable(self):
        class FooSync:
            def __call__(self): ...

        class FooAsync:
            async def __call__(self): ...

        assert not concurrency.is_async(FooSync)
        assert concurrency.is_async(FooAsync)

    def test_partial(self):
        def foo_sync(x: int): ...

        partial_foo_sync = functools.partial(foo_sync, x=1)

        async def foo_async(x: int): ...

        partial_foo_async = functools.partial(foo_async, x=1)

        assert not concurrency.is_async(partial_foo_sync)
        assert concurrency.is_async(partial_foo_async)


class TestCaseRun:
    async def test_async(self):
        async def foo_async(x: int):
            return x + 2

        assert await concurrency.run(foo_async, 3) == 5

    async def test_sync(self):
        def foo_sync(x: int):
            return x * 2

        assert await concurrency.run(foo_sync, 3) == 6


class TestCaseRunTaskGroup:
    async def test_run_task_group(self):
        expected_result = ["foo", "bar"]

        async def foo():
            return "foo"

        async def bar():
            return "bar"

        result = [t.result() for t in await concurrency.run_task_group(foo(), bar())]

        assert result == expected_result


class TestCaseAsyncProcess:
    @pytest.mark.parametrize(
        ["target"],
        (
            pytest.param(sync_event_task, id="sync"),
            pytest.param(async_event_task, id="async"),
            pytest.param(None, id="no_target"),
        ),
    )
    async def test_async_process(self, target):
        event = multiprocessing.Event()

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore")
            process = concurrency.AsyncProcess(target=target, args=(event,))
            process.start()
            process.join()

        if target:
            assert event.wait(5.0)



import typing as t

import pytest

from flama.injection.components import Component
from flama.injection.resolver import Parameter

Foo = t.NewType("Foo", int)


class Bar:
    def __init__(self, x=None):
        self.x = x


class TestCaseComponent:
    @pytest.fixture
    def foo_component(self):
        class FooComponent(Component):
            def resolve(self, x: int) -> Foo:
                return Foo(x)

        return FooComponent()

    @pytest.mark.parametrize(
        ["param_type", "param_name", "expected_id_suffix"],
        (
            pytest.param(int, "x", ":int", id="type_class"),
            pytest.param(Bar(), "x", ":Bar", id="type_object"),
            pytest.param(Parameter, "x", ":Parameter:x", id="parameter"),
        ),
    )
    def test_identity(self, param_type, param_name, expected_id_suffix):
        class BarComponent(Component):
            def resolve(self, x: param_type) -> Bar:
                return Bar(x)

        component = BarComponent()

        assert (
            component.identity(Parameter(param_name, annotation=param_type)) == f"{id(param_type)}{expected_id_suffix}"
        )

    @pytest.mark.parametrize(
        "parameter,expected",
        [
            pytest.param(Parameter("foo", annotation=Foo), True, id="handle"),
            pytest.param(Parameter("foo", annotation=int), False, id="not_handle"),
        ],
    )
    def test_can_handle_parameter(self, foo_component, parameter, expected):
        assert foo_component.can_handle_parameter(parameter) == expected

    def test_signature(self, foo_component):
        assert foo_component.signature() == {"x": Parameter("x", int)}

    async def test_call(self, foo_component):
        assert await foo_component(1) == Foo(1)



import click

from flama.cli.commands.model import command as model_command
from flama.cli.commands.run import command as run_command
from flama.cli.commands.serve import command as serve_command
from flama.cli.commands.start import command as start_command


@click.group()
@click.version_option(
    package_name="Flama",
    message="Flama %(version)s",
    help="Check the version of your locally installed Flama",
)
@click.help_option(help="Get help about how to use Flama CLI")
def cli():
    """
    Fire up your models with Flama 🔥
    """
    ...


cli.add_command(run_command)
cli.add_command(serve_command)
cli.add_command(start_command)
cli.add_command(model_command)


if __name__ == "__main__":
    cli()



import base64
import hashlib
import json
import typing as t

from flama.authentication import exceptions
from flama.authentication.jwt.algorithms import HMACAlgorithm

if t.TYPE_CHECKING:
    from flama.authentication.jwt.algorithms import SignAlgorithm

__all__ = ["JWS"]


class JWS:
    """JSON Web Signature (JWS) implementation.

    It is used to create and decode signed JWT tokens, and to validate the signature of the token. The token is signed
    using the algorithm specified in the header. The supported algorithms are:
    - HMAC with SHA-256
    - HMAC with SHA-384
    - HMAC with SHA-512
    """

    ALGORITHMS = {
        "HS256": HMACAlgorithm(hashlib.sha256),
        "HS384": HMACAlgorithm(hashlib.sha384),
        "HS512": HMACAlgorithm(hashlib.sha512),
    }

    @classmethod
    def _get_algorithm(cls, header: dict[str, t.Any]) -> "SignAlgorithm":
        """Get the algorithm to sign the token.

        It gets the algorithm from the header, and it returns the corresponding algorithm implementation.

        :param header: JWT header.
        :return: Algorithm implementation.
        """
        if "alg" not in header:
            raise exceptions.JWTDecodeException("Missing algorithm in header")

        if header["alg"] not in cls.ALGORITHMS:
            raise exceptions.JWTDecodeException(f"Unsupported algorithm '{header['alg']}'")

        return cls.ALGORITHMS[header["alg"]]

    @classmethod
    def encode(cls, header: dict[str, t.Any], payload: dict[str, t.Any], key: bytes) -> bytes:
        """Encode a JWS token.

        It generates a signed token using the given key. The result is a JWT token with a format of:
        <header>.<payload>.<signature>

        :param header: JWT header.
        :param payload: JWT payload.
        :param key: Key used to sign the token.
        :return: Encoded token.
        """
        header_segment = base64.urlsafe_b64encode(json.dumps(header).encode())
        payload_segment = base64.urlsafe_b64encode(json.dumps(payload).encode())

        algorithm = cls._get_algorithm(header)

        signing_input = b".".join([header_segment, payload_segment])
        signature = base64.urlsafe_b64encode(algorithm.sign(signing_input, key))
        return b".".join([header_segment, payload_segment, signature])

    @classmethod
    def decode(cls, token: bytes, key: bytes) -> tuple[dict[str, t.Any], dict[str, t.Any], bytes]:
        """Decode a JWS token.

        It decode and validate the signature of the token. The token format must be: <header>.<payload>.<signature>

        The header, payload and signature are constructed from the decoded token.

        :param token: Token to decode.
        :param key: Key used to sign the token.
        :return: A tuple with the header, payload and signature of the token.
        :raises JWTDecodeException: If the token format is not correct.
        :raises JWTValidateException: If the token is not valid.
        """
        try:
            signing_input, signature = token.rsplit(b".", 1)
            header_segment, payload_segment = signing_input.split(b".", 1)
        except ValueError:
            raise exceptions.JWTDecodeException("Not enough segments")

        try:
            header = json.loads(base64.urlsafe_b64decode(header_segment))
        except ValueError:
            raise exceptions.JWTDecodeException("Wrong header format")

        try:
            payload = json.loads(base64.urlsafe_b64decode(payload_segment))
        except ValueError:
            raise exceptions.JWTDecodeException("Wrong payload format")

        algorithm = cls._get_algorithm(header)

        if not algorithm.verify(signing_input, base64.urlsafe_b64decode(signature), key):
            raise exceptions.JWTValidateException(f"Signature verification failed for token '{token.decode()}'")

        return header, payload, signature



import pytest

from flama import Flama
from flama.sqlalchemy import SQLAlchemyModule


@pytest.fixture(scope="function")
def app():
    return Flama(schema=None, docs=None, modules={SQLAlchemyModule("sqlite+aiosqlite://")})



import inspect
import logging
import typing as t

from flama import compat, endpoints, exceptions, types, websockets
from flama.routing.routes.base import BaseEndpointWrapper, BaseRoute

__all__ = ["WebSocketRoute"]

logger = logging.getLogger(__name__)


class WebSocketFunctionWrapper(BaseEndpointWrapper):
    async def __call__(self, scope: types.Scope, receive: types.Receive, send: types.Send) -> None:
        """Performs a request.

        :param scope: ASGI scope.
        :param receive: ASGI receive.
        :param send: ASGI send.
        """
        app = scope["app"]
        scope["path"] = scope.get("root_path", "").rstrip("/") + scope["path"]
        scope["root_path"] = ""
        route, route_scope = app.router.resolve_route(scope)
        state = {
            "scope": route_scope,
            "receive": receive,
            "send": send,
            "exc": None,
            "app": app,
            "root_app": scope["root_app"],
            "route": route,
            "websocket": websockets.WebSocket(route_scope, receive, send),
            "websocket_encoding": None,
            "websocket_code": None,
            "websocket_message": None,
        }

        injected_func = await app.injector.inject(self.handler, state)
        await injected_func(**route_scope.get("kwargs", {}))


class WebSocketEndpointWrapper(BaseEndpointWrapper):
    async def __call__(self, scope: types.Scope, receive: types.Receive, send: types.Send) -> None:
        """Performs a request.

        :param scope: ASGI scope.
        :param receive: ASGI receive.
        :param send: ASGI send.
        """
        await self.handler(scope, receive, send)


class WebSocketRoute(BaseRoute):
    def __init__(
        self,
        path: str,
        endpoint: types.WebSocketHandler,
        *,
        name: t.Optional[str] = None,
        include_in_schema: bool = True,
        pagination: t.Optional[types.Pagination] = None,
        tags: t.Optional[dict[str, t.Any]] = None,
    ):
        """A route definition of a websocket endpoint.

        :param path: URL path.
        :param endpoint: Websocket endpoint or function.
        :param name: Route name.
        :param include_in_schema: True if this route must be listed as part of the App schema.
        :param pagination: Apply a pagination technique.
        :param tags: Route tags.
        """

        if not (self.is_endpoint(endpoint) or (not inspect.isclass(endpoint) and callable(endpoint))):
            raise exceptions.ApplicationError("Endpoint must be a callable or a WebSocketEndpoint subclass")

        name = endpoint.__name__ if name is None else name
        wrapper = WebSocketEndpointWrapper if inspect.isclass(endpoint) else WebSocketFunctionWrapper

        super().__init__(
            path, wrapper(endpoint, pagination=pagination), name=name, include_in_schema=include_in_schema, tags=tags
        )

        self.app: BaseEndpointWrapper

    async def __call__(self, scope: types.Scope, receive: types.Receive, send: types.Send) -> None:
        if scope["type"] == "websocket":
            await self.handle(types.Scope({**scope, **self.route_scope(scope)}), receive, send)

    def __eq__(self, other: t.Any) -> bool:
        return super().__eq__(other) and isinstance(other, WebSocketRoute)

    @staticmethod
    def is_endpoint(
        x: types.WebSocketHandler,
    ) -> compat.TypeGuard[type[endpoints.WebSocketEndpoint]]:  # PORT: Replace compat when stop supporting 3.9
        return inspect.isclass(x) and issubclass(x, endpoints.WebSocketEndpoint)

    def endpoint_handlers(self) -> dict[str, t.Callable]:
        """Return a mapping of all possible endpoints of this route.

        Useful to identify all endpoints by HTTP methods.

        :return: Mapping of all endpoints.
        """
        if self.is_endpoint(self.endpoint):
            return self.endpoint.allowed_handlers()

        return {"WEBSOCKET": self.endpoint}

    def match(self, scope: types.Scope) -> BaseRoute.Match:
        """Check if this route matches with given scope.

        :param scope: ASGI scope.
        :return: Match.
        """
        if scope["type"] != "websocket":
            return self.Match.none

        return super().match(scope)



import abc
import asyncio
import inspect
import typing as t

from flama.injection.exceptions import ComponentError, ComponentNotFound
from flama.injection.resolver import Parameter

__all__ = ["Component", "Components"]


class Component(metaclass=abc.ABCMeta):
    def identity(self, parameter: Parameter) -> str:
        """Each component needs a unique identifier string that we use for lookups from the `state` dictionary when we
        run the dependency injection.

        :param parameter: The parameter to check if that component can handle it.
        :return: Unique identifier.
        """
        try:
            parameter_type = parameter.annotation.__name__
        except AttributeError:
            parameter_type = parameter.annotation.__class__.__name__
        component_id = f"{id(parameter.annotation)}:{parameter_type}"

        # If `resolve` includes `Parameter` then use an id that is additionally parameterized by the parameter name.
        args = inspect.signature(self.resolve).parameters.values()  # type: ignore[attr-defined]
        if Parameter in [arg.annotation for arg in args]:
            component_id += f":{parameter.name.lower()}"

        return component_id

    def can_handle_parameter(self, parameter: Parameter) -> bool:
        """The default behavior is for components to handle whatever class is used as the return annotation by the
        `resolve` method.

        You can override this for more customized styles, for example if you wanted name-based parameter resolution, or
        if you want to provide a value for a range of different types.

        :param parameter: The parameter to check if that component can handle it.
        :return: True if this component can handle the given parameter.
        """
        return_annotation = inspect.signature(self.resolve).return_annotation  # type: ignore[attr-defined]
        if return_annotation is inspect.Signature.empty:
            raise ComponentError(
                f"Component '{self.__class__.__name__}' must include a return annotation on the 'resolve' method, or "
                f"override 'can_handle_parameter'"
            )

        return parameter.annotation is return_annotation

    def signature(self) -> dict[str, Parameter]:
        """Component resolver signature.

        :return: Component resolver signature.
        """
        return {
            k: Parameter.from_parameter(v)
            for k, v in inspect.signature(self.resolve).parameters.items()  # type: ignore[attr-defined]
        }

    @property
    def use_parameter(self) -> bool:
        return any(x for x in self.signature().values() if x.annotation is Parameter)

    async def __call__(self, *args, **kwargs):
        """Performs a resolution by calling this component's resolve method.

        :param args: Resolve positional arguments.
        :param kwargs: Resolve keyword arguments.
        :return: Resolve result.
        """
        if asyncio.iscoroutinefunction(self.resolve):  # type: ignore[attr-defined]
            return await self.resolve(*args, **kwargs)  # type: ignore[attr-defined]

        return self.resolve(*args, **kwargs)  # type: ignore[attr-defined]

    def __str__(self) -> str:
        return str(self.__class__.__name__)


class Components(tuple[Component, ...]):
    def __new__(cls, components: t.Optional[t.Union[t.Sequence[Component], set[Component]]] = None):
        return super().__new__(cls, components or [])

    def __eq__(self, other: t.Any) -> bool:
        try:
            return super().__eq__(tuple(other))  # type: ignore[arg-type]
        except TypeError:
            return False

    def find_handler(self, parameter: Parameter) -> Component:
        """Look for a component that can handles given parameter.

        :param parameter: a parameter.
        :return: the component that handles the parameter.
        """
        for component in self:
            if component.can_handle_parameter(parameter):
                return component
        else:
            raise ComponentNotFound(parameter)



import typing as t

from flama import compat, types
from flama.pagination.mixins import LimitOffsetMixin, PageNumberMixin

__all__ = ["paginator"]

P = compat.ParamSpec("P")  # PORT: Replace compat when stop supporting 3.9
R = t.TypeVar("R", covariant=True)


class Paginator(LimitOffsetMixin, PageNumberMixin):
    def __init__(self):
        self.schemas = {}

    def paginate(self, pagination: types.Pagination, func: t.Callable[P, R]) -> t.Callable[P, R]:
        return {
            "limit_offset": self._paginate_limit_offset,
            "page_number": self._paginate_page_number,
        }[pagination](func)


paginator = Paginator()



import typing as t

import pydantic

import flama
from flama import Flama, schemas

app = Flama(
    openapi={
        "info": {
            "title": "Puppy Register",  # API title
            "version": "0.1",  # API version
            "description": "A register of puppies",  # API description
        }
    },
    schema="/schema/",  # Path to expose OpenAPI schema
    docs="/docs/",  # Path to expose Docs application
)


class Puppy(pydantic.BaseModel):
    id: int
    name: str
    age: int

    @pydantic.field_validator("age")
    def minimum_age_validation(cls, v):
        if v < 0:
            raise ValueError("Age must be positive")

        return v


app.schema.register_schema("Puppy", Puppy)


def home():
    return {"hello": "world"}


def list_puppies(name: t.Optional[str] = None) -> t.Annotated[list[schemas.SchemaType], schemas.SchemaMetadata(Puppy)]:
    """
    tags:
        - puppy
    summary:
        List puppies.
    description:
        List the puppies collection. There is an optional query parameter that
        specifies a name for filtering the collection based on it.
    responses:
        200:
            description: List puppies.
    """
    ...


def create_puppy(
    puppy: t.Annotated[schemas.SchemaType, schemas.SchemaMetadata(Puppy)],
) -> t.Annotated[schemas.SchemaType, schemas.SchemaMetadata(Puppy)]:
    """
    tags:
        - puppy
    summary:
        Create a new puppy.
    description:
        Create a new puppy using data validated from request body and add it
        to the collection.
    responses:
        200:
            description: Puppy created successfully.
    """
    ...


app.add_route("/puppy/", list_puppies, methods=["GET"])
app.add_route("/puppy/", create_puppy, methods=["POST"])

if __name__ == "__main__":
    flama.run(flama_app=app, server_host="0.0.0.0", server_port=8080)



from unittest.mock import Mock

import pytest

from flama.models import ModelComponent
from flama.models.models.pytorch import PyTorchModel
from flama.models.models.sklearn import SKLearnModel
from flama.models.models.tensorflow import TensorFlowModel


@pytest.fixture(params=["tensorflow", "sklearn", "torch"])
def model(request):
    return {
        "sklearn": SKLearnModel(Mock(), Mock(), Mock()),
        "tensorflow": TensorFlowModel(Mock(), Mock(), Mock()),
        "torch": PyTorchModel(Mock(), Mock(), Mock()),
    }[request.param]


@pytest.fixture
def component(model):
    class SpecificModelComponent(ModelComponent):
        def resolve(self) -> type(model):
            return self.model

    return SpecificModelComponent(model)



import dataclasses
import datetime
import logging
import typing as t
from http.cookies import SimpleCookie

from flama import Flama, types
from flama.authentication.types import AccessToken, RefreshToken
from flama.exceptions import HTTPException
from flama.http import Request as HTTPRequest

logger = logging.getLogger(__name__)

__all__ = ["Endpoint", "Authentication", "Request", "Response", "Error", "TelemetryData"]


@dataclasses.dataclass
class Endpoint:
    path: str
    name: t.Optional[str]
    tags: dict[str, t.Any]

    @classmethod
    async def from_scope(cls, *, scope: types.Scope, receive: types.Receive, send: types.Send) -> "Endpoint":
        app: Flama = scope["app"]
        route, _ = app.router.resolve_route(scope)

        return cls(path=str(route.path), name=route.name, tags=route.tags)

    def to_dict(self) -> dict[str, t.Any]:
        return {
            "path": self.path,
            "name": self.name,
            "tags": self.tags,
        }


@dataclasses.dataclass
class Authentication:
    access: t.Optional[AccessToken]
    refresh: t.Optional[RefreshToken]

    @classmethod
    async def from_scope(cls, *, scope: types.Scope, receive: types.Receive, send: types.Send) -> "Authentication":
        app: Flama = scope["app"]
        context = {"scope": scope, "request": HTTPRequest(scope, receive=receive)}

        try:
            access = await app.injector.resolve(AccessToken).value(context)
        except Exception:
            access = None

        try:
            refresh = await app.injector.resolve(RefreshToken).value(context)
        except Exception:
            refresh = None

        return cls(access=access, refresh=refresh)

    def to_dict(self) -> dict[str, t.Any]:
        return {
            "access": self.access.to_dict() if self.access else None,
            "refresh": self.refresh.to_dict() if self.refresh else None,
        }


@dataclasses.dataclass
class Request:
    headers: dict[str, t.Any]
    cookies: dict[str, t.Any]
    query_parameters: dict[str, t.Any]
    path_parameters: dict[str, t.Any]
    body: bytes = b""
    timestamp: datetime.datetime = dataclasses.field(
        init=False, default_factory=lambda: datetime.datetime.now(datetime.timezone.utc)
    )

    @classmethod
    async def from_scope(cls, *, scope: types.Scope, receive: types.Receive, send: types.Send) -> "Request":
        app: Flama = scope["app"]
        context = {"scope": scope, "request": HTTPRequest(scope, receive=receive), "route": app.resolve_route(scope)[0]}

        headers = dict(await app.injector.resolve(types.Headers).value(context))
        cookies = dict(await app.injector.resolve(types.Cookies).value(context))
        query = dict(await app.injector.resolve(types.QueryParams).value(context))
        path = dict(await app.injector.resolve(types.PathParams).value(context))

        return cls(headers=headers, cookies=cookies, query_parameters=query, path_parameters=path)

    def to_dict(self) -> dict[str, t.Any]:
        return {
            "timestamp": self.timestamp.isoformat(),
            "headers": self.headers,
            "cookies": self.cookies,
            "query_parameters": self.query_parameters,
            "path_parameters": self.path_parameters,
            "body": self.body,
        }


@dataclasses.dataclass
class Response:
    headers: t.Optional[dict[str, t.Any]]
    cookies: t.Optional[dict[str, t.Any]] = dataclasses.field(init=False)
    body: bytes = b""
    status_code: t.Optional[int] = None
    timestamp: datetime.datetime = dataclasses.field(
        init=False, default_factory=lambda: datetime.datetime.now(datetime.timezone.utc)
    )

    def __post_init__(self):
        if self.headers:
            cookie = SimpleCookie()
            cookie.load(self.headers.get("cookie", ""))
        else:
            cookie = {}

        self.cookies = {
            str(name): {**{str(k): str(v) for k, v in morsel.items()}, "value": morsel.value}
            for name, morsel in cookie.items()
        }

    def to_dict(self) -> dict[str, t.Any]:
        return {
            "timestamp": self.timestamp.isoformat(),
            "headers": self.headers,
            "cookies": self.cookies,
            "body": self.body,
            "status_code": self.status_code,
        }


@dataclasses.dataclass
class Error:
    detail: str
    status_code: t.Optional[int] = None
    timestamp: datetime.datetime = dataclasses.field(
        init=False, default_factory=lambda: datetime.datetime.now(datetime.timezone.utc)
    )

    @classmethod
    async def from_exception(cls, *, exception: Exception) -> "Error":
        if isinstance(exception, HTTPException):
            return cls(status_code=exception.status_code, detail=str(exception.detail))

        return cls(detail=str(exception))

    def to_dict(self) -> dict[str, t.Any]:
        return {
            "timestamp": self.timestamp.isoformat(),
            "detail": self.detail,
            "status_code": self.status_code,
        }


@dataclasses.dataclass
class TelemetryData:
    type: t.Literal["http", "websocket"]
    endpoint: Endpoint
    authentication: Authentication
    request: Request
    response: t.Optional[Response] = None
    error: t.Optional[Error] = None
    extra: dict[t.Any, t.Any] = dataclasses.field(default_factory=dict)

    @classmethod
    async def from_scope(cls, *, scope: types.Scope, receive: types.Receive, send: types.Send) -> "TelemetryData":
        return cls(
            type=scope["type"],
            endpoint=await Endpoint.from_scope(scope=scope, receive=receive, send=send),
            authentication=await Authentication.from_scope(scope=scope, receive=receive, send=send),
            request=await Request.from_scope(scope=scope, receive=receive, send=send),
        )

    def to_dict(self) -> dict[str, t.Any]:
        return {
            "type": self.type,
            "endpoint": self.endpoint.to_dict(),
            "authentication": self.authentication.to_dict(),
            "request": self.request.to_dict(),
            "response": self.response.to_dict() if self.response else None,
            "error": self.error.to_dict() if self.error else None,
            "extra": self.extra,
        }



import typing as t

from pydantic import BaseModel, Field

__all__ = [
    "APIError",
    "DropCollection",
    "LimitOffsetMeta",
    "LimitOffset",
    "PageNumberMeta",
    "PageNumber",
    "MLModelInput",
    "MLModelOutput",
    "SCHEMAS",
]


class APIError(BaseModel):
    status_code: int = Field(title="status_code", description="HTTP status code")
    detail: t.Union[str, dict[str, t.Any]] = Field(title="detail", description="Error detail")
    error: t.Optional[str] = Field(title="type", description="Exception or error type")


class DropCollection(BaseModel):
    deleted: int = Field(title="deleted", description="Number of deleted elements")


class LimitOffsetMeta(BaseModel):
    limit: int = Field(title="limit", description="Number of retrieved items")
    offset: int = Field(title="offset", description="Collection offset")
    count: t.Optional[int] = Field(title="count", description="Total number of items")


class LimitOffset(BaseModel):
    meta: LimitOffsetMeta = Field(title="meta", description="Pagination metadata")
    data: list[t.Any] = Field(title="data", description="Paginated data")


class PageNumberMeta(BaseModel):
    page: int = Field(title="page", description="Current page number")
    page_size: int = Field(title="page_size", description="Page size")
    count: t.Optional[int] = Field(title="count", description="Total number of items")


class PageNumber(BaseModel):
    meta: PageNumberMeta = Field(title="meta", description="Pagination metadata")
    data: list[t.Any] = Field(title="data", description="Paginated data")


class MLModelInput(BaseModel):
    input: list[t.Any] = Field(title="input", description="Model input")


class MLModelOutput(BaseModel):
    output: list[t.Any] = Field(title="output", description="Model output")


SCHEMAS = {
    "flama.APIError": APIError,
    "flama.DropCollection": DropCollection,
    "flama.LimitOffsetMeta": LimitOffsetMeta,
    "flama.LimitOffset": LimitOffset,
    "flama.PageNumberMeta": PageNumberMeta,
    "flama.PageNumber": PageNumber,
    "flama.MLModelInput": MLModelInput,
    "flama.MLModelOutput": MLModelOutput,
}
